import logging

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _
from rest_framework.exceptions import ValidationError

from waldur_core.cost_tracking import signals as cost_signals
from waldur_core.core import utils as core_utils

from . import models, log, tasks

logger = logging.getLogger(__name__)


def log_invoice_state_transition(sender, instance, created=False, **kwargs):
    if created:
        return

    state = instance.state
    if state == models.Invoice.States.PENDING or state == instance.tracker.previous('state'):
        return

    if state == models.Invoice.States.CREATED:
        log.event_logger.invoice.info(
            'Invoice for customer {customer_name} has been created.',
            event_type='invoice_created',
            event_context={'month': instance.month, 'year': instance.year, 'customer': instance.customer}
        )
    elif state == models.Invoice.States.PAID:
        log.event_logger.invoice.info(
            'Invoice for customer {customer_name} has been paid.',
            event_type='invoice_paid',
            event_context={'month': instance.month, 'year': instance.year, 'customer': instance.customer}
        )
    elif state == models.Invoice.States.CANCELED:
        log.event_logger.invoice.info(
            'Invoice for customer {customer_name} has been canceled.',
            event_type='invoice_canceled',
            event_context={'month': instance.month, 'year': instance.year, 'customer': instance.customer}
        )


def set_tax_percent_on_invoice_creation(sender, instance, **kwargs):
    if instance.pk is not None:
        return

    instance.tax_percent = instance.customer.default_tax_percent


def set_project_name_on_invoice_item_creation(sender, instance, created=False, **kwargs):
    if created and instance.project:
        item = instance
        item.project_name = item.project.name
        item.project_uuid = item.project.uuid.hex
        item.save(update_fields=('project_name', 'project_uuid'))


def update_invoice_item_on_project_name_update(sender, instance, **kwargs):
    project = instance

    if not project.tracker.has_changed('name'):
        return

    query = Q(project=project, invoice__state=models.Invoice.States.PENDING)
    for item in models.InvoiceItem.objects.filter(query).only('pk'):
        item.project_name = project.name
        item.save(update_fields=['project_name'])


def emit_invoice_created_event(sender, instance, created=False, **kwargs):
    if created:
        return

    state = instance.state
    if state != models.Invoice.States.CREATED or state == instance.tracker.previous('state'):
        return

    cost_signals.invoice_created.send(sender=models.Invoice,
                                      invoice=instance,
                                      issuer_details=settings.WALDUR_INVOICES['ISSUER_DETAILS'])


def prevent_deletion_of_customer_with_invoice(sender, instance, user, **kwargs):
    if user.is_staff:
        return
    PENDING = models.Invoice.States.PENDING
    for invoice in models.Invoice.objects.filter(customer=instance):
        if invoice.state != PENDING or invoice.price > 0:
            raise ValidationError(_('Can\'t delete organization with invoice %s.') % invoice)


def update_current_cost_when_invoice_item_is_updated(sender, instance, created=False, **kwargs):
    invoice_item = instance
    if created or set(invoice_item.tracker.changed()) & {'start', 'end', 'quantity', 'unit_price'}:
        transaction.on_commit(lambda: invoice_item.invoice.update_current_cost())


def update_current_cost_when_invoice_item_is_deleted(sender, instance, **kwargs):
    def update_invoice():
        try:
            instance.invoice.update_current_cost()
        except ObjectDoesNotExist:
            # It is okay to skip cache invalidation if invoice has been already removed
            pass

    transaction.on_commit(update_invoice)


@transaction.atomic()
def adjust_openstack_items_for_downtime(downtime):
    items = models.InvoiceItem.objects.filter(
        downtime.get_intersection_subquery(),
        scope=downtime.package,
    )

    for item in items:
        # outside
        if downtime.start <= item.start and item.end <= downtime.end:
            item.create_compensation(item.name, start=item.start, end=item.end)

        # inside
        elif item.start <= downtime.start and downtime.end <= item.end:
            item.create_compensation(item.name, start=downtime.start, end=downtime.end)

        # left
        elif downtime.end >= item.start and downtime.end <= item.end:
            item.create_compensation(item.name, start=item.start, end=downtime.end)

        # right
        elif downtime.start >= item.start and downtime.start <= item.end:
            item.create_compensation(item.name, start=downtime.start, end=item.end)


def adjust_invoice_items_for_downtime(sender, instance, created=False, **kwargs):
    downtime = instance
    if not created:
        logger.warning('Invoice items are not adjusted when downtime record is changed. '
                       'Record ID: %s', downtime.id)

    adjust_openstack_items_for_downtime(downtime)


def update_invoice_pdf(sender, instance, created=False, **kwargs):
    if created:
        return

    invoice = instance

    if not invoice.tracker.has_changed('current_cost'):
        return

    serialized_invoice = core_utils.serialize_instance(invoice)
    tasks.create_invoice_pdf.delay(serialized_invoice)



import logging
from urllib.parse import urlparse

from django.core.validators import RegexValidator
from django.db import models
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker
from model_utils.models import TimeStampedModel

from waldur_core.core import models as core_models
from waldur_core.core.fields import JSONField
from waldur_core.logging.loggers import LoggableMixin
from waldur_core.quotas import models as quotas_models, fields as quotas_fields
from waldur_core.structure import models as structure_models, utils as structure_utils
from waldur_geo_ip.utils import get_coordinates_by_ip
from waldur_openstack.openstack import models as openstack_models
from waldur_openstack.openstack_base import models as openstack_base_models

logger = logging.getLogger(__name__)


TenantQuotas = openstack_models.Tenant.Quotas


class OpenStackTenantService(structure_models.Service):
    projects = models.ManyToManyField(
        structure_models.Project, related_name='openstack_tenant_services', through='OpenStackTenantServiceProjectLink')

    class Meta:
        unique_together = ('customer', 'settings')
        verbose_name = _('OpenStackTenant provider')
        verbose_name_plural = _('OpenStackTenant providers')

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant'


class OpenStackTenantServiceProjectLink(structure_models.CloudServiceProjectLink):
    service = models.ForeignKey(on_delete=models.CASCADE, to=OpenStackTenantService)

    class Meta(structure_models.CloudServiceProjectLink.Meta):
        verbose_name = _('OpenStackTenant provider project link')
        verbose_name_plural = _('OpenStackTenant provider project links')

    class Quotas(quotas_models.QuotaModelMixin.Quotas):
        vcpu = quotas_fields.TotalQuotaField(
            target_models=lambda: [Instance],
            path_to_scope='service_project_link',
            target_field='cores',
        )
        ram = quotas_fields.TotalQuotaField(
            target_models=lambda: [Instance],
            path_to_scope='service_project_link',
            target_field='ram',
        )
        storage = quotas_fields.TotalQuotaField(
            target_models=lambda: [Volume, Snapshot],
            path_to_scope='service_project_link',
            target_field='size',
        )

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-spl'


class Flavor(LoggableMixin, structure_models.ServiceProperty):
    cores = models.PositiveSmallIntegerField(help_text=_('Number of cores in a VM'))
    ram = models.PositiveIntegerField(help_text=_('Memory size in MiB'))
    disk = models.PositiveIntegerField(help_text=_('Root disk size in MiB'))

    class Meta:
        unique_together = ('settings', 'backend_id')

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-flavor'

    @classmethod
    def get_backend_fields(cls):
        return super(Flavor, cls).get_backend_fields() + ('cores', 'ram', 'disk')


class Image(openstack_base_models.BaseImage):
    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-image'


class SecurityGroup(core_models.DescribableMixin, structure_models.ServiceProperty):

    class Meta:
        unique_together = ('settings', 'backend_id')

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-sgp'


class SecurityGroupRule(openstack_base_models.BaseSecurityGroupRule):
    security_group = models.ForeignKey(on_delete=models.CASCADE, to=SecurityGroup, related_name='rules')


class TenantQuotaMixin(quotas_models.SharedQuotaMixin):
    """
    It allows to update both service settings and shared tenant quotas.
    """

    def get_quota_scopes(self):
        service_settings = self.service_project_link.service.settings
        return service_settings, service_settings.scope


class FloatingIP(structure_models.ServiceProperty):
    address = models.GenericIPAddressField(protocol='IPv4', null=True, default=None)
    runtime_state = models.CharField(max_length=30)
    backend_network_id = models.CharField(max_length=255, editable=False)
    is_booked = models.BooleanField(default=False,
                                    help_text=_('Marks if floating IP has been booked for provisioning.'))
    internal_ip = models.ForeignKey('InternalIP', related_name='floating_ips', null=True, on_delete=models.SET_NULL)
    tracker = FieldTracker()

    class Meta:
        unique_together = ('settings', 'address')
        verbose_name = _('Floating IP')
        verbose_name_plural = _('Floating IPs')

    def __str__(self):
        return '%s:%s | %s' % (self.address, self.runtime_state, self.settings)

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-fip'

    def get_backend(self):
        return self.settings.get_backend()

    def increase_backend_quotas_usage(self, validate=True):
        self.settings.add_quota_usage(self.settings.Quotas.floating_ip_count, 1, validate=validate)

    def decrease_backend_quotas_usage(self):
        self.settings.add_quota_usage(self.settings.Quotas.floating_ip_count, -1)

    @classmethod
    def get_backend_fields(cls):
        return super(FloatingIP, cls).get_backend_fields() + ('address', 'runtime_state', 'backend_network_id')


class Volume(TenantQuotaMixin, structure_models.Volume):
    # backend_id is nullable on purpose, otherwise
    # it wouldn't be possible to put a unique constraint on it
    backend_id = models.CharField(max_length=255, blank=True, null=True)

    service_project_link = models.ForeignKey(
        OpenStackTenantServiceProjectLink, related_name='volumes', on_delete=models.PROTECT)
    instance = models.ForeignKey(on_delete=models.CASCADE, to='Instance', related_name='volumes', blank=True, null=True)
    device = models.CharField(
        max_length=50, blank=True,
        validators=[RegexValidator('^/dev/[a-zA-Z0-9]+$',
                                   message=_('Device should match pattern "/dev/alphanumeric+"'))],
        help_text=_('Name of volume as instance device e.g. /dev/vdb.'))
    bootable = models.BooleanField(default=False)
    metadata = JSONField(blank=True)
    image = models.ForeignKey(Image, blank=True, null=True, on_delete=models.SET_NULL)
    image_name = models.CharField(max_length=150, blank=True)
    image_metadata = JSONField(blank=True)
    type = models.ForeignKey('VolumeType', blank=True, null=True, on_delete=models.SET_NULL)
    availability_zone = models.ForeignKey('VolumeAvailabilityZone', blank=True, null=True, on_delete=models.SET_NULL)
    source_snapshot = models.ForeignKey('Snapshot', related_name='volumes', blank=True, null=True,
                                        on_delete=models.SET_NULL)
    # TODO: Move this fields to resource model.
    action = models.CharField(max_length=50, blank=True)
    action_details = JSONField(default=dict)

    tracker = FieldTracker()

    class Meta:
        unique_together = ('service_project_link', 'backend_id')

    def get_quota_deltas(self):
        return {
            TenantQuotas.volumes: 1,
            TenantQuotas.volumes_size: self.size,
            TenantQuotas.storage: self.size,
        }

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-volume'

    @classmethod
    def get_backend_fields(cls):
        return super(Volume, cls).get_backend_fields() + ('name', 'description', 'size', 'metadata', 'type', 'bootable',
                                                          'runtime_state', 'device', 'instance', 'availability_zone',
                                                          'image')


class Snapshot(TenantQuotaMixin, structure_models.Snapshot):
    # backend_id is nullable on purpose, otherwise
    # it wouldn't be possible to put a unique constraint on it
    backend_id = models.CharField(max_length=255, blank=True, null=True)

    service_project_link = models.ForeignKey(
        OpenStackTenantServiceProjectLink, related_name='snapshots', on_delete=models.PROTECT)
    source_volume = models.ForeignKey(Volume, related_name='snapshots', null=True, on_delete=models.PROTECT)
    metadata = JSONField(blank=True)
    # TODO: Move this fields to resource model.
    action = models.CharField(max_length=50, blank=True)
    action_details = JSONField(default=dict)
    snapshot_schedule = models.ForeignKey('SnapshotSchedule',
                                          blank=True,
                                          null=True,
                                          on_delete=models.SET_NULL,
                                          related_name='snapshots')

    tracker = FieldTracker()

    kept_until = models.DateTimeField(
        null=True,
        blank=True,
        help_text=_('Guaranteed time of snapshot retention. If null - keep forever.'))

    class Meta:
        unique_together = ('service_project_link', 'backend_id')

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-snapshot'

    def get_quota_deltas(self):
        return {
            TenantQuotas.snapshots: 1,
            TenantQuotas.snapshots_size: self.size,
            TenantQuotas.storage: self.size,
        }

    @classmethod
    def get_backend_fields(cls):
        return super(Snapshot, cls).get_backend_fields() + ('name', 'description', 'size', 'metadata', 'source_volume',
                                                            'runtime_state')


class SnapshotRestoration(core_models.UuidMixin, TimeStampedModel):
    snapshot = models.ForeignKey(on_delete=models.CASCADE, to=Snapshot, related_name='restorations')
    volume = models.OneToOneField(Volume, related_name='restoration', on_delete=models.CASCADE)

    class Permissions:
        customer_path = 'snapshot__service_project_link__project__customer'
        project_path = 'snapshot__service_project_link__project'


class InstanceAvailabilityZone(structure_models.BaseServiceProperty):
    settings = models.ForeignKey(on_delete=models.CASCADE, to=structure_models.ServiceSettings, related_name='+')
    available = models.BooleanField(default=True)

    class Meta:
        unique_together = ('settings', 'name')

    def __str__(self):
        return self.name

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-instance-availability-zone'


class Instance(TenantQuotaMixin, structure_models.VirtualMachine):

    class RuntimeStates:
        # All possible OpenStack Instance states on backend.
        # See https://docs.openstack.org/developer/nova/vmstates.html
        ACTIVE = 'ACTIVE'
        BUILDING = 'BUILDING'
        DELETED = 'DELETED'
        SOFT_DELETED = 'SOFT_DELETED'
        ERROR = 'ERROR'
        UNKNOWN = 'UNKNOWN'
        HARD_REBOOT = 'HARD_REBOOT'
        REBOOT = 'REBOOT'
        REBUILD = 'REBUILD'
        PASSWORD = 'PASSWORD'
        PAUSED = 'PAUSED'
        RESCUED = 'RESCUED'
        RESIZED = 'RESIZED'
        REVERT_RESIZE = 'REVERT_RESIZE'
        SHUTOFF = 'SHUTOFF'
        STOPPED = 'STOPPED'
        SUSPENDED = 'SUSPENDED'
        VERIFY_RESIZE = 'VERIFY_RESIZE'

    # backend_id is nullable on purpose, otherwise
    # it wouldn't be possible to put a unique constraint on it
    backend_id = models.CharField(max_length=255, blank=True, null=True)
    service_project_link = models.ForeignKey(
        OpenStackTenantServiceProjectLink, related_name='instances', on_delete=models.PROTECT)

    availability_zone = models.ForeignKey(InstanceAvailabilityZone, blank=True, null=True, on_delete=models.SET_NULL)
    flavor_name = models.CharField(max_length=255, blank=True)
    flavor_disk = models.PositiveIntegerField(default=0, help_text=_('Flavor disk size in MiB'))
    security_groups = models.ManyToManyField(SecurityGroup, related_name='instances')
    # TODO: Move this fields to resource model.
    action = models.CharField(max_length=50, blank=True)
    action_details = JSONField(default=dict)
    subnets = models.ManyToManyField('SubNet', through='InternalIP')

    tracker = FieldTracker()

    class Meta:
        unique_together = ('service_project_link', 'backend_id')

    @property
    def external_ips(self):
        return list(self.floating_ips.values_list('address', flat=True))

    @property
    def internal_ips(self):
        return list(self.internal_ips_set.values_list('ip4_address', flat=True))

    @property
    def size(self):
        return self.volumes.aggregate(models.Sum('size'))['size__sum']

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-instance'

    def get_log_fields(self):
        return ('uuid', 'name', 'type', 'service_project_link', 'ram', 'cores',)

    def detect_coordinates(self):
        settings = self.service_project_link.service.settings
        options = settings.options or {}
        if 'latitude' in options and 'longitude' in options:
            return structure_utils.Coordinates(latitude=settings['latitude'], longitude=settings['longitude'])
        else:
            hostname = urlparse(settings.backend_url).hostname
            if hostname:
                return get_coordinates_by_ip(hostname)

    def get_quota_deltas(self):
        return {
            TenantQuotas.instances: 1,
            TenantQuotas.ram: self.ram,
            TenantQuotas.vcpu: self.cores,
        }

    @property
    def floating_ips(self):
        return FloatingIP.objects.filter(internal_ip__instance=self)

    @classmethod
    def get_backend_fields(cls):
        return super(Instance, cls).get_backend_fields() + ('flavor_name', 'flavor_disk', 'ram', 'cores', 'disk',
                                                            'runtime_state', 'availability_zone')

    @classmethod
    def get_online_state(cls):
        return Instance.RuntimeStates.ACTIVE

    @classmethod
    def get_offline_state(cls):
        return Instance.RuntimeStates.SHUTOFF


class Backup(structure_models.SubResource):
    service_project_link = models.ForeignKey(
        OpenStackTenantServiceProjectLink, related_name='backups', on_delete=models.PROTECT)
    instance = models.ForeignKey(Instance, related_name='backups', on_delete=models.PROTECT)
    backup_schedule = models.ForeignKey('BackupSchedule', blank=True, null=True,
                                        on_delete=models.SET_NULL,
                                        related_name='backups')
    kept_until = models.DateTimeField(
        null=True,
        blank=True,
        help_text=_('Guaranteed time of backup retention. If null - keep forever.'))
    metadata = JSONField(
        blank=True,
        help_text=_('Additional information about backup, can be used for backup restoration or deletion'),
    )
    snapshots = models.ManyToManyField('Snapshot', related_name='backups')

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-backup'


class BackupRestoration(core_models.UuidMixin, TimeStampedModel):
    """ This model corresponds to instance restoration from backup. """
    backup = models.ForeignKey(on_delete=models.CASCADE, to=Backup, related_name='restorations')
    instance = models.OneToOneField(Instance, related_name='+', on_delete=models.CASCADE)
    flavor = models.ForeignKey(Flavor, related_name='+', null=True, blank=True, on_delete=models.SET_NULL)

    class Permissions:
        customer_path = 'backup__service_project_link__project__customer'
        project_path = 'backup__service_project_link__project'


class BaseSchedule(structure_models.NewResource, core_models.ScheduleMixin):
    retention_time = models.PositiveIntegerField(
        help_text=_('Retention time in days, if 0 - resource will be kept forever'))
    maximal_number_of_resources = models.PositiveSmallIntegerField()
    call_count = models.PositiveSmallIntegerField(default=0,
                                                  help_text=_('How many times a resource schedule was called.'))

    class Meta:
        abstract = True


class BackupSchedule(BaseSchedule):
    service_project_link = models.ForeignKey(
        OpenStackTenantServiceProjectLink, related_name='backup_schedules', on_delete=models.PROTECT)
    instance = models.ForeignKey(on_delete=models.CASCADE, to=Instance, related_name='backup_schedules')

    tracker = FieldTracker()

    def __str__(self):
        return 'BackupSchedule of %s. Active: %s' % (self.instance, self.is_active)

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-backup-schedule'


class SnapshotSchedule(BaseSchedule):
    service_project_link = models.ForeignKey(
        OpenStackTenantServiceProjectLink, related_name='snapshot_schedules', on_delete=models.PROTECT)
    source_volume = models.ForeignKey(on_delete=models.CASCADE, to=Volume, related_name='snapshot_schedules')

    tracker = FieldTracker()

    def __str__(self):
        return 'SnapshotSchedule of %s. Active: %s' % (self.source_volume, self.is_active)

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-snapshot-schedule'


class Network(core_models.DescribableMixin, structure_models.ServiceProperty):
    is_external = models.BooleanField(default=False)
    type = models.CharField(max_length=50, blank=True)
    segmentation_id = models.IntegerField(null=True)

    class Meta:
        unique_together = ('settings', 'backend_id')

    def __str__(self):
        return self.name

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-network'


class SubNet(core_models.DescribableMixin, structure_models.ServiceProperty):
    network = models.ForeignKey(on_delete=models.CASCADE, to=Network, related_name='subnets')
    cidr = models.CharField(max_length=32, blank=True)
    gateway_ip = models.GenericIPAddressField(protocol='IPv4', null=True)
    allocation_pools = JSONField(default=dict)
    ip_version = models.SmallIntegerField(default=4)
    enable_dhcp = models.BooleanField(default=True)
    dns_nameservers = JSONField(default=list, help_text=_('List of DNS name servers associated with the subnet.'))

    class Meta:
        verbose_name = _('Subnet')
        verbose_name_plural = _('Subnets')
        unique_together = ('settings', 'backend_id')

    def __str__(self):
        return '%s (%s)' % (self.name, self.cidr)

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-subnet'


class InternalIP(openstack_base_models.Port):
    """
    Instance may have several IP addresses in the same subnet
    if shared IPs are implemented using Virtual Router Redundancy Protocol.
    """
    # Name "internal_ips" is reserved by virtual machine mixin and corresponds to list of internal IPs.
    # So another related name should be used.
    instance = models.ForeignKey(on_delete=models.CASCADE, to=Instance, related_name='internal_ips_set', null=True)
    subnet = models.ForeignKey(on_delete=models.CASCADE, to=SubNet, related_name='internal_ips')

    # backend_id is nullable on purpose, otherwise
    # it wouldn't be possible to put a unique constraint on it
    backend_id = models.CharField(max_length=255, null=True)
    settings = models.ForeignKey(on_delete=models.CASCADE, to=structure_models.ServiceSettings, related_name='+')
    tracker = FieldTracker()

    class Meta:
        unique_together = ('backend_id', 'settings')


class VolumeType(openstack_base_models.BaseVolumeType):

    is_default = models.BooleanField(default=False)

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-volume-type'


class VolumeAvailabilityZone(structure_models.BaseServiceProperty):
    settings = models.ForeignKey(on_delete=models.CASCADE, to=structure_models.ServiceSettings, related_name='+')
    available = models.BooleanField(default=True)

    class Meta:
        unique_together = ('settings', 'name')

    def __str__(self):
        return self.name

    @classmethod
    def get_url_name(cls):
        return 'openstacktenant-volume-availability-zone'



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-06-19 13:10
import django.core.validators
from django.db import migrations, models
import re
import waldur_freeipa.models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_freeipa', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='profile',
            name='username',
            field=models.CharField(help_text='Letters, numbers and ./+/-/_ characters', max_length=32, unique=True, validators=[waldur_freeipa.models.validate_username, django.core.validators.RegexValidator(re.compile(r'^[a-zA-Z0-9_.][a-zA-Z0-9_.-]*[a-zA-Z0-9_.$-]?$'), 'Enter a valid username.', 'invalid')], verbose_name='username'),
        ),
    ]



import base64
from collections import OrderedDict
from datetime import timedelta
import logging

from django.core.exceptions import ImproperlyConfigured, MultipleObjectsReturned, ObjectDoesNotExist
from django.urls import reverse, Resolver404
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers
from rest_framework.fields import Field, ReadOnlyField
from rest_framework.exceptions import ValidationError

from waldur_core.core import utils as core_utils
from waldur_core.core.fields import TimestampField
from waldur_core.core.signals import pre_serializer_fields

logger = logging.getLogger(__name__)


class AuthTokenSerializer(serializers.Serializer):
    """
    API token serializer loosely based on DRF's default AuthTokenSerializer.
    but with the logic of authorization is extracted to view.
    """
    # Fields are both required, non-blank and don't allow nulls by default
    username = serializers.CharField()
    password = serializers.CharField()


class Base64Field(serializers.CharField):
    def to_internal_value(self, data):
        value = super(Base64Field, self).to_internal_value(data)
        try:
            base64.b64decode(value)
            return value
        except (TypeError, ValueError):
            raise serializers.ValidationError(_('This field should a be valid Base64 encoded string.'))

    def to_representation(self, value):
        value = super(Base64Field, self).to_representation(value)
        if isinstance(value, str):
            value = value.encode('utf-8')
        return base64.b64encode(value)


class BasicInfoSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        fields = ('url', 'uuid', 'name')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class UnboundSerializerMethodField(ReadOnlyField):
    """
    A field that gets its value by calling a provided filter callback.
    """

    def __init__(self, filter_function, *args, **kwargs):
        self.filter_function = filter_function
        super(UnboundSerializerMethodField, self).__init__(*args, **kwargs)

    def to_representation(self, value):
        request = self.context.get('request')
        return self.filter_function(value, request)


class GenericRelatedField(Field):
    """
    A custom field to use for the `tagged_object` generic relationship.
    """
    read_only = False
    _default_view_name = '%(model_name)s-detail'
    lookup_fields = ['uuid', 'pk']

    def __init__(self, related_models=(), **kwargs):
        super(GenericRelatedField, self).__init__(**kwargs)
        self._related_models = related_models

    @property
    def related_models(self):
        val = self._related_models
        return callable(val) and val() or val

    def _get_url(self, obj):
        """
        Gets object url
        """
        format_kwargs = {
            'app_label': obj._meta.app_label,
        }
        try:
            format_kwargs['model_name'] = getattr(obj.__class__, 'get_url_name')()
        except AttributeError:
            format_kwargs['model_name'] = obj._meta.object_name.lower()
        return self._default_view_name % format_kwargs

    def _get_request(self):
        try:
            return self.context['request']
        except KeyError:
            raise AttributeError('GenericRelatedField have to be initialized with `request` in context')

    def to_representation(self, obj):
        """
        Serializes any object to his url representation
        """
        kwargs = None
        for field in self.lookup_fields:
            if hasattr(obj, field):
                kwargs = {field: getattr(obj, field)}
                break
        if kwargs is None:
            raise AttributeError('Related object does not have any of lookup_fields')
        if self.related_models and not isinstance(obj, tuple(self.related_models)):
            return None
        request = self._get_request()
        return request.build_absolute_uri(reverse(self._get_url(obj), kwargs=kwargs))

    def to_internal_value(self, data):
        """
        Restores model instance from its url
        """
        if not data:
            return None
        request = self._get_request()
        user = request.user
        try:
            obj = core_utils.instance_from_url(data, user=user)
            model = obj.__class__
        except ValueError:
            raise serializers.ValidationError(_('URL is invalid: %s.') % data)
        except (Resolver404, AttributeError, MultipleObjectsReturned, ObjectDoesNotExist):
            raise serializers.ValidationError(_("Can't restore object from url: %s") % data)

        if self.related_models and model not in self.related_models:
            context = (model, ', '.join(str(model) for model in self.related_models))
            message = _('%s is not valid. Valid models are: %s') % context
            raise serializers.ValidationError(message)

        return obj


class AugmentedSerializerMixin:
    """
    This mixin provides several extensions to stock Serializer class:

    1.  Add extra fields to serializer from dependent applications in a way
        that doesn't introduce circular dependencies.

        To achieve this, dependent application should subscribe
        to pre_serializer_fields signal and inject additional fields.

        Example of signal handler implementation:

        from waldur_core.structure.serializers import CustomerSerializer

        def add_customer_name(sender, fields, **kwargs):
            fields['customer_name'] = ReadOnlyField(source='customer.name')

        pre_serializer_fields.connect(
            handlers.add_customer_name,
            sender=CustomerSerializer
        )

    2.  Declaratively add attributes fields of related entities for ModelSerializers.

        To achieve list related fields whose attributes you want to include.

        Example:
            class ProjectSerializer(AugmentedSerializerMixin,
                                    serializers.HyperlinkedModelSerializer):
                class Meta:
                    model = models.Project
                    fields = (
                        'url', 'uuid', 'name',
                        'customer', 'customer_uuid', 'customer_name',
                    )
                    related_paths = ('customer',)

            # This is equivalent to listing the fields explicitly,
            # by default "uuid" and "name" fields of related object are added:

            class ProjectSerializer(AugmentedSerializerMixin,
                                    serializers.HyperlinkedModelSerializer):
                customer_uuid = serializers.ReadOnlyField(source='customer.uuid')
                customer_name = serializers.ReadOnlyField(source='customer.name')
                class Meta:
                    model = models.Project
                    fields = (
                        'url', 'uuid', 'name',
                        'customer', 'customer_uuid', 'customer_name',
                    )
                    lookup_field = 'uuid'

            # The fields of related object can be customized:

            class ProjectSerializer(AugmentedSerializerMixin,
                                    serializers.HyperlinkedModelSerializer):
                class Meta:
                    model = models.Project
                    fields = (
                        'url', 'uuid', 'name',
                        'customer', 'customer_uuid',
                        'customer_name', 'customer_native_name',
                    )
                    related_paths = {
                        'customer': ('uuid', 'name', 'native_name')
                    }

    3.  Protect some fields from change.

        Example:
            class ProjectSerializer(AugmentedSerializerMixin,
                                    serializers.HyperlinkedModelSerializer):
                class Meta:
                    model = models.Project
                    fields = ('url', 'uuid', 'name', 'customer')
                    protected_fields = ('customer',)

    4. This mixin overrides "get_extra_kwargs" method and puts "view_name" to extra_kwargs
    or uses URL name specified in a model of serialized object.
    """

    def get_fields(self):
        fields = super(AugmentedSerializerMixin, self).get_fields()
        pre_serializer_fields.send(sender=self.__class__, fields=fields)

        try:
            protected_fields = self.Meta.protected_fields
        except AttributeError:
            pass
        else:
            try:
                method = self.context['view'].request.method
            except (KeyError, AttributeError):
                return fields

            if method in ('PUT', 'PATCH'):
                for field in protected_fields:
                    fields[field].read_only = True

        return fields

    def _get_related_paths(self):
        try:
            related_paths = self.Meta.related_paths
        except AttributeError:
            return {}

        if not isinstance(self, serializers.ModelSerializer):
            raise ImproperlyConfigured(
                'related_paths can be defined only for ModelSerializer.'
            )

        if isinstance(related_paths, (list, tuple)):
            related_paths = {path: ('name', 'uuid') for path in related_paths}

        return related_paths

    def build_unknown_field(self, field_name, model_class):
        related_paths = self._get_related_paths()

        related_field_source_map = {
            '{0}_{1}'.format(path.split('.')[-1], attribute): '{0}.{1}'.format(path, attribute)
            for path, attributes in related_paths.items()
            for attribute in attributes
        }

        try:
            return serializers.ReadOnlyField, {'source': related_field_source_map[field_name]}
        except KeyError:
            return super(AugmentedSerializerMixin, self).build_unknown_field(field_name, model_class)

    def get_extra_kwargs(self):
        extra_kwargs = super(AugmentedSerializerMixin, self).get_extra_kwargs()

        if hasattr(self.Meta, 'view_name'):
            view_name = self.Meta.view_name
        else:
            view_name = core_utils.get_detail_view_name(self.Meta.model)

        if 'url' in extra_kwargs:
            extra_kwargs['url']['view_name'] = view_name
        else:
            extra_kwargs['url'] = {'view_name': view_name}

        return extra_kwargs


class RestrictedSerializerMixin:
    """
    This mixin allows to specify list of fields to be rendered by serializer.
    It expects that request is available in serializer's context.
    """

    FIELDS_PARAM_NAME = 'field'

    def get_fields(self):
        fields = super(RestrictedSerializerMixin, self).get_fields()
        if 'request' not in self.context:
            return fields
        query_params = self.context['request'].query_params
        keys = query_params.getlist(self.FIELDS_PARAM_NAME)
        keys = set(key for key in keys if key in fields.keys())
        if not keys:
            return fields
        return OrderedDict(((key, value) for key, value in fields.items() if key in keys))


class RequiredFieldsMixin:
    """
    This mixin allows to specify list of required fields.
    It expects list of field names as Meta.required_fields attribute.
    """

    def get_fields(self):
        fields = super(RequiredFieldsMixin, self).get_fields()
        required_fields = getattr(self.Meta, 'required_fields') or []
        for name in required_fields:
            field = fields.get(name)
            if field:
                field.required = True
        return fields


class ExtraFieldOptionsMixin:
    """
    This mixin allows to specify extra fields metadata.
    It expects dictionary of field name and options as Meta.extra_field_options attribute.
    """

    def get_fields(self):
        fields = super(ExtraFieldOptionsMixin, self).get_fields()
        extra_field_options = getattr(self.Meta, 'extra_field_options', {})
        for name, options in extra_field_options.items():
            field = fields.get(name)
            if field:
                for key, val in options.items():
                    setattr(field, key, val)
        return fields


class HyperlinkedRelatedModelSerializer(serializers.HyperlinkedModelSerializer):
    def __init__(self, **kwargs):
        self.queryset = kwargs.pop('queryset', None)
        assert self.queryset is not None or kwargs.get('read_only', None), (
            'Relational field must provide a `queryset` argument, '
            'or set read_only=`True`.'
        )
        assert not (self.queryset is not None and kwargs.get('read_only', None)), (
            'Relational fields should not provide a `queryset` argument, '
            'when setting read_only=`True`.'
        )
        super(HyperlinkedRelatedModelSerializer, self).__init__(**kwargs)

    def to_internal_value(self, data):
        if 'url' not in data:
            raise serializers.ValidationError(_('URL has to be defined for related object.'))
        url_field = self.fields['url']

        # This is tricky: self.fields['url'] is the one generated
        # based on Meta.fields.
        # By default ModelSerializer generates it as HyperlinkedIdentityField,
        # which is read-only, thus it doesn't get deserialized from POST body.
        # So, we "borrow" its view_name and lookup_field to create
        # a HyperlinkedRelatedField which can turn url into a proper model
        # instance.
        url = serializers.HyperlinkedRelatedField(
            queryset=self.queryset.all(),
            view_name=url_field.view_name,
            lookup_field=url_field.lookup_field,
        )

        return url.to_internal_value(data['url'])


class TimestampIntervalSerializer(serializers.Serializer):
    start = TimestampField(required=False)
    end = TimestampField(required=False)

    def validate(self, data):
        """
        Check that the start is before the end.
        """
        if 'start' in data and 'end' in data and data['start'] >= data['end']:
            raise serializers.ValidationError(_('End must occur after start.'))
        return data

    # TimeInterval serializer is used for validation only. We are providing custom method for such serializers
    # to avoid confusion with to_internal_value or to_representation DRF methods.
    def get_filter_data(self):
        """ Return start and end as datetime """
        return self.validated_data


class HistorySerializer(serializers.Serializer):
    """
    Receive datetime as timestamps and converts them to list of datetimes

    Support 2 types of input data:
     - start, end and points_count - interval from <start> to <end> will be automatically split into
                                     <points_count> pieces
     - point_list - list of timestamps that will be converted to datetime points

    """
    start = TimestampField(required=False)
    end = TimestampField(required=False)
    points_count = serializers.IntegerField(min_value=2, required=False)
    point_list = serializers.ListField(
        child=TimestampField(),
        required=False
    )

    def validate(self, attrs):
        autosplit_fields = {'start', 'end', 'points_count'}
        if ('point_list' not in attrs or not attrs['point_list']) and not autosplit_fields == set(attrs.keys()):
            raise serializers.ValidationError(
                _('Not enough parameters for historical data. '
                  '(Either "point" or "start" + "end" + "points_count" parameters have to be provided).'))
        if 'point_list' in attrs and autosplit_fields & set(attrs.keys()):
            raise serializers.ValidationError(
                _('Too many parameters for historical data. '
                  '(Either "point" or "start" + "end" + "points_count" parameters have to be provided).'))
        if 'point_list' not in attrs and not attrs['start'] < attrs['end']:
            raise serializers.ValidationError(_('Start timestamps have to be later than end timestamps.'))
        return attrs

    # History serializer is used for validation only. We are providing custom method for such serializers
    # to avoid confusion with to_internal_value or to_representation DRF methods.
    def get_filter_data(self):
        if 'point_list' in self.validated_data:
            return self.validated_data['point_list']
        else:
            interval = ((self.validated_data['end'] - self.validated_data['start']) /
                        (self.validated_data['points_count'] - 1))
            return [self.validated_data['start'] + interval * i for i in range(self.validated_data['points_count'])]


class TimelineSerializer(serializers.Serializer):

    INTERVAL_CHOICES = ('hour', 'day', 'week', 'month')

    start_time = TimestampField(default=lambda: core_utils.timeshift(days=-1))
    end_time = TimestampField(default=lambda: core_utils.timeshift())
    interval = serializers.ChoiceField(choices=INTERVAL_CHOICES, default='day')

    def get_date_points(self):
        start_time = self.validated_data['start_time']
        end_time = self.validated_data['end_time']
        interval = self.validated_data['interval']

        if interval == 'hour':
            start_point = start_time.replace(second=0, minute=0, microsecond=0)
            interval = timedelta(hours=1)
        elif interval == 'day':
            start_point = start_time.replace(hour=0, second=0, minute=0, microsecond=0)
            interval = timedelta(days=1)
        elif interval == 'week':
            start_point = start_time.replace(hour=0, second=0, minute=0, microsecond=0)
            interval = timedelta(days=7)
        elif interval == 'month':
            start_point = start_time.replace(hour=0, second=0, minute=0, microsecond=0)
            interval = timedelta(days=30)

        points = [start_time]
        current_point = start_point
        while current_point <= end_time:
            points.append(current_point)
            current_point += interval
        if points[-1] != end_time:
            points.append(end_time)

        return [p for p in points if start_time <= p <= end_time]


class BaseSummarySerializer(serializers.Serializer):
    """ Serializer that renders each instance with its own specific serializer """

    @classmethod
    def get_serializer(cls, model):
        raise NotImplementedError('Method `get_serializer` should be implemented for SummarySerializer.')

    @classmethod
    def eager_load(cls, summary_queryset, request):
        optimized_querysets = []
        for queryset in summary_queryset.querysets:
            serializer = cls.get_serializer(queryset.model)
            optimized_querysets.append(serializer.eager_load(queryset, request))
        summary_queryset.querysets = optimized_querysets
        return summary_queryset

    def to_representation(self, instance):
        serializer = self.get_serializer(instance.__class__)
        return serializer(instance, context=self.context).data


class GeoLocationField(serializers.JSONField):
    def __init__(self, *args, **kwargs):
        validators = kwargs.get('validators', [])

        def geo_location_validator(value):
            if value is not None:
                if not isinstance(value, list):
                    raise ValidationError(_('GeoLocationField should be a list of dictionaries.'))
                else:
                    for location in value:
                        if not isinstance(location, dict):
                            raise ValidationError(_('GeoLocationField should be a list of dictionaries.'))
                        if not {'latitude', 'longitude'}.issubset(location.keys()):
                            raise ValidationError(_('GeoLocationField should be a list of dictionaries. For example: '
                                                    '[{"latitude": 123, "longitude": 345}, '
                                                    '{"latitude": 456, "longitude": 678}]'))
            return value

        validators.append(geo_location_validator)
        super(GeoLocationField, self).__init__(validators=validators, *args, **kwargs)


class UnicodeIntegerField(serializers.IntegerField):

    def to_internal_value(self, data):
        if isinstance(data, str):
            data = core_utils.normalize_unicode(data)
        return super(UnicodeIntegerField, self).to_internal_value(data)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-09-12 15:22
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_slurm', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='allocation',
            name='gpu_limit',
            field=models.IntegerField(default=-1),
        ),
        migrations.AddField(
            model_name='allocation',
            name='gpu_usage',
            field=models.IntegerField(default=0),
        ),
        migrations.AddField(
            model_name='allocation',
            name='ram_limit',
            field=models.IntegerField(default=-1),
        ),
        migrations.AddField(
            model_name='allocation',
            name='ram_usage',
            field=models.IntegerField(default=0),
        ),
    ]



from django.conf import settings as conf_settings
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.utils.functional import cached_property

from waldur_core.core import utils as core_utils, models as core_models
from waldur_core.structure import ServiceBackend
from waldur_core.structure.models import ServiceSettings
from waldur_core.structure.utils import update_pulled_fields
from waldur_mastermind.common.utils import parse_datetime

from . import models, signals, client


class RancherBackend(ServiceBackend):

    DEFAULTS = {
        'cloud_init_template':
            '#cloud-config\n'
            'packages: \n'
            '  - curl\n'
            'runcmd:\n'
            '  - curl -fsSL https://get.docker.com -o get-docker.sh; sh get-docker.sh\n'
            '  - sudo systemctl start docker\n'
            '  - sudo systemctl enable docker\n'
            '  - [ sh, -c, "{command}" ]\n'
    }

    def __init__(self, settings):
        """
        :type settings: :class:`waldur_core.structure.models.ServiceSettings`
        """
        self.settings = settings

    @cached_property
    def client(self):
        """
        Construct Rancher REST API client using credentials specified in the service settings.
        """
        rancher_client = client.RancherClient(self.host, verify_ssl=False)
        rancher_client.login(self.settings.username, self.settings.password)
        return rancher_client

    @cached_property
    def host(self):
        return self.settings.backend_url.strip('/')

    def pull_service_properties(self):
        self.pull_catalogs()

    def get_kubeconfig_file(self, cluster):
        return self.client.get_kubeconfig_file(cluster.backend_id)

    def create_cluster(self, cluster):
        backend_cluster = self.client.create_cluster(cluster.name)
        self._backend_cluster_to_cluster(backend_cluster, cluster)
        self.client.create_cluster_registration_token(cluster.backend_id)
        cluster.node_command = self.client.get_node_command(cluster.backend_id)
        cluster.save()

    def delete_cluster(self, cluster):
        if cluster.backend_id:
            self.client.delete_cluster(cluster.backend_id)

        cluster.delete()

    def delete_node(self, node):
        if node.backend_id:
            self.client.delete_node(node.backend_id)

        node.delete()

    def update_cluster(self, cluster):
        backend_cluster = self._cluster_to_backend_cluster(cluster)
        self.client.update_cluster(cluster.backend_id, backend_cluster)

    def _backend_cluster_to_cluster(self, backend_cluster, cluster):
        cluster.backend_id = backend_cluster['id']
        cluster.name = backend_cluster['name']
        cluster.runtime_state = backend_cluster['state']

    def _cluster_to_backend_cluster(self, cluster):
        return {'name': cluster.name}

    def _backend_node_to_node(self, backend_node):
        return {'backend_id': backend_node['nodeId'], 'name': backend_node['hostnameOverride']}

    def get_clusters_for_import(self):
        cur_clusters = set(models.Cluster.objects.filter(service_project_link__service__settings=self.settings)
                           .values_list('backend_id', flat=True))
        backend_clusters = self.client.list_clusters()
        return filter(lambda c: c['id'] not in cur_clusters, backend_clusters)

    def import_cluster(self, backend_id, service_project_link):
        backend_cluster = self.client.get_cluster(backend_id)
        cluster = models.Cluster(
            backend_id=backend_id,
            service_project_link=service_project_link,
            state=models.Cluster.States.OK,
            runtime_state=backend_cluster['state'])
        self.pull_cluster(cluster, backend_cluster)
        return cluster

    def pull_cluster(self, cluster, backend_cluster=None):
        backend_cluster = backend_cluster or self.client.get_cluster(cluster.backend_id)
        self._backend_cluster_to_cluster(backend_cluster, cluster)
        cluster.save()
        backend_nodes = backend_cluster.get('appliedSpec', {}).get('rancherKubernetesEngineConfig', {}).get('nodes', [])

        for backend_node in backend_nodes:
            roles = backend_node.get('role', [])

            # If the node has not been requested from Waldur, so it will be created
            node, created = models.Node.objects.get_or_create(
                name=backend_node.get('hostnameOverride'),
                cluster=cluster,
                defaults=dict(
                    state=models.Node.States.OK,
                    backend_id=backend_node.get('nodeId'),
                    controlplane_role='controlplane' in roles,
                    etcd_role='etcd' in roles,
                    worker_role='worker' in roles
                )
            )

            if not node.backend_id:
                # If the node has been requested from Waldur, but it has not been synchronized
                node.state = models.Node.States.OK
                node.backend_id = backend_node.get('nodeId')
                node.controlplane_role = 'controlplane' in roles
                node.etcd_role = 'etcd' in roles
                node.worker_role = 'worker' in roles
                node.save()

            # Update details in all cases.
            self.update_node_details(node)

    def pull_cluster_runtime_state(self, cluster):
        backend_cluster = self.client.get_cluster(cluster.backend_id)
        self._backend_cluster_to_cluster(backend_cluster, cluster)
        cluster.save()

    def check_cluster_creating(self, cluster):
        self.pull_cluster_runtime_state(cluster)

        if cluster.runtime_state == conf_settings.WALDUR_RANCHER['ACTIVE_CLUSTER_STATE']:
            cluster.state = models.Cluster.States.OK
            cluster.save()
            return

        for node in cluster.node_set.filter(Q(controlplane_role=True) | Q(etcd_role=True)):
            controlplane_role = etcd_role = False
            if node.instance.state not in [core_models.StateMixin.States.ERRED,
                                           core_models.StateMixin.States.DELETING,
                                           core_models.StateMixin.States.DELETION_SCHEDULED]:
                if node.controlplane_role:
                    controlplane_role = True
                if node.etcd_role:
                    etcd_role = True
                if controlplane_role and etcd_role:
                    # We make a return if one or more VMs with 'controlplane' and 'etcd' roles exist
                    # and they haven't a state 'error' or 'delete'.
                    # Here 'return' means that cluster state checking must be retry later.
                    return

        cluster.error_message = 'The cluster is not connected with any ' \
                                'non-failed VM\'s with \'controlplane\' or \'etcd\' roles.'
        cluster.runtime_state = 'error'
        cluster.state = core_models.StateMixin.States.ERRED
        cluster.save()

    def get_cluster_nodes(self, backend_id):
        backend_cluster = self.client.get_cluster(backend_id)
        nodes = backend_cluster.get('appliedSpec', {}).get('rancherKubernetesEngineConfig', {}).get('nodes', [])
        return [self._backend_node_to_node(node) for node in nodes]

    def node_is_active(self, backend_id):
        backend_node = self.client.get_node(backend_id)
        return backend_node['state'] == conf_settings.WALDUR_RANCHER['ACTIVE_NODE_STATE']

    def update_node_details(self, node):
        if not node.backend_id:
            return

        backend_node = self.client.get_node(node.backend_id)

        # rancher can skip return of some fields when node is being created,
        # so avoid crashing by supporting missing values
        def get_backend_node_field(*args):
            value = backend_node

            for arg in args:
                if isinstance(value, dict):
                    value = value.get(arg)
                else:
                    return

            return value

        def update_node_field(*args, field):
            value = get_backend_node_field(*args)
            if value:
                setattr(node, field, value)

        update_node_field('labels', field='labels')
        update_node_field('annotations', field='annotations')
        update_node_field('info', 'os', 'dockerVersion', field='docker_version')
        update_node_field('info', 'kubernetes', 'kubeletVersion', field='k8s_version')
        cpu_allocated = get_backend_node_field('requested', 'cpu')

        if cpu_allocated:
            node.cpu_allocated = \
                core_utils.parse_int(cpu_allocated) / 1000  # convert data from 380m to 0.38

        ram_allocated = get_backend_node_field('requested', 'memory')
        update_node_field('allocatable', 'cpu', field='cpu_total')

        if ram_allocated:
            node.ram_allocated = \
                int(core_utils.parse_int(ram_allocated) / 2 ** 20)  # convert data to Mi

        ram_total = get_backend_node_field('allocatable', 'memory')

        if ram_total:
            node.ram_total = \
                int(core_utils.parse_int(ram_total) / 2 ** 20)  # convert data to Mi

        update_node_field('requested', 'pods', field='pods_allocated')
        update_node_field('allocatable', 'pods', field='pods_total')
        update_node_field('state', field='runtime_state')

        if node.runtime_state == conf_settings.WALDUR_RANCHER['ACTIVE_NODE_STATE']:
            node.state = models.Node.States.OK
        else:
            node.state = models.Node.States.ERRED
        return node.save()

    def create_user(self, user):
        if user.backend_id:
            return

        password = models.RancherUser.make_random_password()
        response = self.client.create_user(
            name=user.user.username,
            username=user.user.username,
            password=password
        )
        user_id = response['id']
        user.backend_id = user_id
        user.save()
        self.client.create_global_role(user.backend_id, client.GlobalRoleId.user_base)
        signals.rancher_user_created.send(
            sender=models.RancherUser,
            instance=user,
            password=password,
        )

    def delete_user(self, user):
        if user.backend_id:
            self.client.delete_user(user_id=user.backend_id)

        user.delete()

    def block_user(self, user):
        if user.is_active:
            self.client.disable_user(user.backend_id)
            user.is_active = False
            user.save()

    def activate_user(self, user):
        if not user.is_active:
            self.client.enable_user(user.backend_id)
            user.is_active = True
            user.save()

    def create_cluster_role(self, link):
        role = None

        if link.role == models.ClusterRole.CLUSTER_OWNER:
            role = client.ClusterRoleId.cluster_owner

        if link.role == models.ClusterRole.CLUSTER_MEMBER:
            role = client.ClusterRoleId.cluster_member

        response = self.client.create_cluster_role(
            link.user.backend_id,
            link.cluster.backend_id,
            role)
        link_id = response['id']
        link.backend_id = link_id
        link.save()

    def delete_cluster_role(self, link):
        if link.backend_id:
            self.client.delete_cluster_role(cluster_role_id=link.backend_id)

        link.delete()

    def pull_catalogs(self):
        self.pull_global_catalogs()
        self.pull_cluster_catalogs()
        # TODO: Implement Rancher project catalogs synchronization

    def pull_global_catalogs(self):
        remote_catalogs = self.client.list_global_catalogs()
        self.pull_catalogs_for_scope(remote_catalogs, self.settings)

    def pull_cluster_catalogs(self):
        remote_catalogs = self.client.list_cluster_catalogs()
        for cluster in models.Cluster.objects.filter(service_project_link__service__settings=self.settings):
            self.pull_catalogs_for_scope(remote_catalogs, cluster)

    def pull_catalogs_for_scope(self, remote_catalogs, scope):
        content_type = ContentType.objects.get_for_model(scope)
        local_catalogs = models.Catalog.objects.filter(
            content_type=content_type,
            object_id=scope.id,
        )

        remote_catalog_map = {
            catalog['id']: self.remote_catalog_to_local(catalog, content_type, scope.id)
            for catalog in remote_catalogs
        }
        local_catalog_map = {
            catalog.id: catalog
            for catalog in local_catalogs
        }
        remote_catalog_ids = set(remote_catalog_map.keys())
        local_catalog_ids = set(local_catalog_map.keys())

        stale_catalogs = local_catalog_ids - remote_catalog_ids

        new_catalogs = [
            remote_catalog_map[catalog_id]
            for catalog_id in remote_catalog_ids - local_catalog_ids
        ]

        existing_catalogs = remote_catalog_ids & local_catalog_ids
        pulled_fields = {
            'name', 'description', 'catalog_url', 'branch',
            'commit', 'username', 'password', 'runtime_state',
        }
        for catalog_id in existing_catalogs:
            local_catalog = local_catalog_map[catalog_id]
            remote_catalog = remote_catalog_map[catalog_id]
            update_pulled_fields(local_catalog, remote_catalog, pulled_fields)

        models.Catalog.objects.bulk_create(new_catalogs)
        local_catalogs.filter(id__in=stale_catalogs).delete()

    def remote_catalog_to_local(self, remote_catalog, content_type, object_id):
        return models.Catalog(
            content_type=content_type,
            object_id=object_id,
            backend_id=remote_catalog['id'],
            name=remote_catalog['name'],
            description=remote_catalog['description'],
            created=parse_datetime(remote_catalog['created']),
            catalog_url=remote_catalog['url'],
            branch=remote_catalog['branch'],
            commit=remote_catalog.get('commit', ''),
            username=remote_catalog.get('username', ''),
            password=remote_catalog.get('password', ''),
            runtime_state=remote_catalog['state'],
        )

    def refresh_catalog(self, catalog):
        if isinstance(catalog.scope, ServiceSettings):
            return self.client.refresh_global_catalog(catalog.backend_id)
        elif isinstance(catalog.scope, models.Cluster):
            return self.client.refresh_cluster_catalog(catalog.backend_id)
        else:
            return self.client.refresh_project_catalog(catalog.backend_id)

    def delete_catalog(self, catalog):
        if isinstance(catalog.scope, ServiceSettings):
            return self.client.delete_global_catalog(catalog.backend_id)
        elif isinstance(catalog.scope, models.Cluster):
            return self.client.delete_cluster_catalog(catalog.backend_id)
        else:
            return self.client.delete_project_catalog(catalog.backend_id)

    def get_catalog_spec(self, catalog):
        spec = {
            'name': catalog.name,
            'description': catalog.description,
            'url': catalog.catalog_url,
            'branch': catalog.branch,
        }
        if catalog.username:
            spec['username'] = catalog.username
        if catalog.password:
            spec['password'] = catalog.password
        return spec

    def create_catalog(self, catalog):
        spec = self.get_catalog_spec(catalog)

        if isinstance(catalog.scope, ServiceSettings):
            remote_catalog = self.client.create_global_catalog(spec)
        elif isinstance(catalog.scope, models.Cluster):
            spec['clusterId'] = catalog.scope.backend_id
            remote_catalog = self.client.create_cluster_catalog(spec)
        else:
            spec['projectId'] = catalog.scope.backend_id
            remote_catalog = self.client.create_project_catalog(spec)

        catalog.backend_id = remote_catalog['id']
        catalog.runtime_state = remote_catalog['state']
        catalog.save()

    def update_catalog(self, catalog):
        spec = self.get_catalog_spec(catalog)
        if isinstance(catalog.scope, ServiceSettings):
            return self.client.update_global_catalog(catalog.backend_id, spec)
        elif isinstance(catalog.scope, models.Cluster):
            return self.client.update_cluster_catalog(catalog.backend_id, spec)
        else:
            return self.client.update_project_catalog(catalog.backend_id, spec)



from celery import chain

from waldur_core.core import executors as core_executors
from waldur_core.core import tasks as core_tasks
from waldur_openstack.openstack_tenant import executors as openstack_executors
from waldur_openstack.openstack_tenant import models as openstack_models


class RunJobExecutor(core_executors.CreateExecutor):
    @classmethod
    def get_task_signature(cls, job, serialized_job, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_job, 'run_job', state_transition='begin_creating')


class DeleteJobExecutor(core_executors.DeleteExecutor):
    @classmethod
    def get_task_signature(cls, job, serialized_job, **kwargs):
        deletion_tasks = [
            core_tasks.StateTransitionTask().si(serialized_job, state_transition='begin_deleting')
        ]
        for resource in job.get_related_resources():
            force = resource.state == openstack_models.Instance.States.ERRED
            deletion_tasks.append(openstack_executors.InstanceDeleteExecutor.as_signature(
                resource, force=force, delete_volumes=True))
        return chain(*deletion_tasks)



# Generated by Django 2.2.7 on 2019-11-12 19:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('marketplace', '0003_componentusage_recurring'),
    ]

    operations = [
        migrations.AddField(
            model_name='offeringcomponent',
            name='content_type',
            field=models.ForeignKey(null=True, blank=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType'),
        ),
        migrations.AddField(
            model_name='offeringcomponent',
            name='object_id',
            field=models.PositiveIntegerField(null=True, blank=True),
        ),
    ]



import yaml
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.utils.translation import ugettext_lazy as _
import logging
from rest_framework import serializers

from waldur_core.core.models import User
from waldur_core.quotas import exceptions as quotas_exceptions
from waldur_core.structure.models import ProjectRole, ServiceSettings
from waldur_openstack.openstack_tenant import models as openstack_tenant_models
from waldur_rancher.backend import RancherBackend

from . import models, exceptions

logger = logging.getLogger(__name__)


def get_unique_node_name(name, instance_spl, cluster_spl, existing_names=None):
    existing_names = existing_names or []
    # This has a potential risk of race condition when requests to create nodes come exactly at the same time.
    # But we consider this use case highly unrealistic and avoid creation of additional complexity
    # to protect against it
    names_instances = openstack_tenant_models.Instance.objects.filter(service_project_link=instance_spl)\
        .values_list('name', flat=True)
    names_nodes = models.Node.objects.filter(cluster__service_project_link=cluster_spl).values_list('name', flat=True)
    names = list(names_instances) + list(names_nodes) + existing_names

    i = 1
    new_name = '%s-%s' % (name, i)

    while new_name in names:
        i += 1
        new_name = '%s-%s' % (name, i)

    return new_name


def expand_added_nodes(nodes, rancher_spl, tenant_settings, cluster_name):
    project = rancher_spl.project

    try:
        tenant_spl = openstack_tenant_models.OpenStackTenantServiceProjectLink.objects.get(
            project=project,
            service__settings=tenant_settings)
    except ObjectDoesNotExist:
        raise serializers.ValidationError(
            'Service project link for service %s and project %s is not found.' % (
                tenant_settings.name, project.name
            ))

    try:
        base_image_name = rancher_spl.service.settings.get_option('base_image_name')
        image = openstack_tenant_models.Image.objects.get(
            name=base_image_name,
            settings=tenant_settings)
    except ObjectDoesNotExist:
        raise serializers.ValidationError('No matching image found.')

    try:
        group = openstack_tenant_models.SecurityGroup.objects.get(
            name='default',
            settings=tenant_settings)
    except ObjectDoesNotExist:
        raise serializers.ValidationError('Default security group is not found.')

    for node in nodes:
        memory = node.pop('memory', None)
        cpu = node.pop('cpu', None)
        subnet = node.pop('subnet')
        flavor = node.pop('flavor', None)
        roles = node.pop('roles')
        system_volume_size = node.pop('system_volume_size', None)
        system_volume_type = node.pop('system_volume_type', None)
        data_volumes = node.pop('data_volumes', [])

        if subnet.settings != tenant_settings:
            raise serializers.ValidationError(
                'Subnet %s should belong to the service settings %s.' % (
                    subnet.name, tenant_settings.name,
                ))

        validate_data_volumes(data_volumes, tenant_settings)
        flavor = validate_flavor(flavor, roles, tenant_settings, cpu, memory)

        node['initial_data'] = {
            'flavor': flavor.uuid.hex,
            'vcpu': flavor.cores,
            'ram': flavor.ram,
            'image': image.uuid.hex,
            'subnet': subnet.uuid.hex,
            'tenant_service_project_link': tenant_spl.id,
            'group': group.uuid.hex,
            'system_volume_size': system_volume_size,
            'system_volume_type': system_volume_type and system_volume_type.uuid.hex,
            'data_volumes': [{
                'size': volume['size'],
                'volume_type': volume.get('volume_type') and volume.get('volume_type').uuid.hex,
            } for volume in data_volumes]
        }

        if 'controlplane' in list(roles):
            node['controlplane_role'] = True
        if 'etcd' in list(roles):
            node['etcd_role'] = True
        if 'worker' in list(roles):
            node['worker_role'] = True

        node['name'] = get_unique_node_name(cluster_name + '-rancher-node', tenant_spl, rancher_spl,
                                            existing_names=[n['name'] for n in nodes if n.get('name')])

    validate_quotas(nodes, tenant_spl)


def validate_data_volumes(data_volumes, tenant_settings):
    for volume in data_volumes:
        volume_type = volume.get('volume_type')
        if volume_type and volume_type.settings != tenant_settings:
            raise serializers.ValidationError(
                'Volume type %s should belong to the service settings %s.' % (
                    volume_type.name, tenant_settings.name,
                ))

    mount_points = [volume['mount_point'] for volume in data_volumes]
    if len(set(mount_points)) != len(mount_points):
        raise serializers.ValidationError('Each mount point can be specified once at most.')


def validate_flavor(flavor, roles, tenant_settings, cpu=None, memory=None):
    if flavor:
        if cpu or memory:
            raise serializers.ValidationError('Either flavor or cpu and memory should be specified.')
    else:
        if not cpu or not memory:
            raise serializers.ValidationError('Either flavor or cpu and memory should be specified.')

    if not flavor:
        flavor = openstack_tenant_models.Flavor.objects.filter(
            cores__gte=cpu,
            ram__gte=memory,
            settings=tenant_settings). \
            order_by('cores', 'ram').first()

    if not flavor:
        raise serializers.ValidationError('No matching flavor found.')

    if flavor.settings != tenant_settings:
        raise serializers.ValidationError(
            'Flavor %s should belong to the service settings %s.' % (
                flavor.name, tenant_settings.name,
            ))

    requirements = list(filter(lambda x: x[0] in list(roles),
                               settings.WALDUR_RANCHER['ROLE_REQUIREMENT'].items()))
    if requirements:
        cpu_requirements = max([t[1]['CPU'] for t in requirements])
        ram_requirements = max([t[1]['RAM'] for t in requirements])
        if flavor.cores < cpu_requirements:
            raise serializers.ValidationError('Flavor %s does not meet requirements. CPU requirement is %s'
                                              % (flavor, cpu_requirements))
        if flavor.ram < ram_requirements:
            raise serializers.ValidationError('Flavor %s does not meet requirements. RAM requirement is %s'
                                              % (flavor, ram_requirements))

    return flavor


def validate_quotas(nodes, tenant_spl):
    quota_sources = [
        tenant_spl,
        tenant_spl.project,
        tenant_spl.customer,
        tenant_spl.service,
        tenant_spl.service.settings
    ]
    for quota_name in ['storage', 'vcpu', 'ram']:
        requested = sum(get_node_quota(quota_name, node) for node in nodes)

        for source in quota_sources:
            try:
                quota = source.quotas.get(name=quota_name)
                if quota.limit != -1 and (quota.usage + requested > quota.limit):
                    raise quotas_exceptions.QuotaValidationError(
                        _('"%(name)s" quota is over limit. Required: %(usage)s, limit: %(limit)s.') % dict(
                            name=quota_name, usage=quota.usage + requested, limit=quota.limit))
            except ObjectDoesNotExist:
                pass


def get_node_quota(quota_name, node):
    conf = node['initial_data']
    if quota_name == 'storage':
        data_volumes = conf.get('data_volumes', [])
        return conf['system_volume_size'] + sum(volume['size'] for volume in data_volumes)
    else:
        return conf[quota_name]


def format_disk_id(index):
    return '/dev/vd' + (chr(ord('a') + index))


def format_node_command(node):
    roles_command = []

    if node.controlplane_role:
        roles_command.append('--controlplane')

    if node.etcd_role:
        roles_command.append('--etcd')

    if node.worker_role:
        roles_command.append('--worker')

    return node.cluster.node_command + ' ' + ' '.join(roles_command)


def format_node_cloud_config(node):
    node_command = format_node_command(node)
    config_template = node.service_project_link.service.settings.get_option('cloud_init_template')
    user_data = config_template.format(command=node_command)
    data_volumes = node.initial_data.get('data_volumes')

    if data_volumes:
        data_volumes = sorted(data_volumes)
        conf = yaml.parse(user_data)

        # First volume is reserved for system volume, other volumes are data volumes

        conf['mounts'] = [
            [format_disk_id(index + 1), volume['mount_point']]
            for index, volume in enumerate(data_volumes)
        ]

        conf['fs_setup'] = [
            {'device': format_disk_id(index + 1), 'filesystem': 'ext4'}
            for index, volume in enumerate(data_volumes)
        ]
        user_data = yaml.dump(conf)

    return user_data


class SyncUser:
    @staticmethod
    def get_users():
        result = {}

        def add_to_result():
            if user not in result.keys():
                result[user] = {}

            if service_settings not in result[user].keys():
                result[user][service_settings] = []

            result[user][service_settings].append([cluster, role])

        for cluster in models.Cluster.objects.all():
            service_settings = cluster.service_project_link.service.settings
            project = cluster.service_project_link.project
            users = project.get_users()
            owners = project.customer.get_owners()

            for user in users:
                role = 'manager' if project.has_user(user, ProjectRole.MANAGER) \
                    else 'admin' if project.has_user(user, ProjectRole.ADMINISTRATOR) else None
                add_to_result()

            for user in owners:
                role = 'owner'
                add_to_result()

        return result

    @staticmethod
    def create_users(add_users):
        count_created = 0
        count_activated = 0
        for user in add_users:
            for service_settings in add_users[user].keys():
                try:
                    with transaction.atomic():
                        rancher_user, created = models.RancherUser.objects.get_or_create(
                            settings=service_settings,
                            user=user,
                        )
                        backend = RancherBackend(service_settings)

                        if created:
                            backend.create_user(rancher_user)
                            count_created += 1
                        else:
                            backend.activate_user(rancher_user)
                            count_activated += 1
                except exceptions.RancherException as e:
                    logger.error('Error creating or activating user %s. %s' % (user, e))

        return count_created, count_activated

    @staticmethod
    def block_users(rancher_users):
        count = 0

        for user in rancher_users:
            try:
                with transaction.atomic():
                    backend = RancherBackend(user.settings)
                    backend.block_user(user)
                    count += 1
            except exceptions.RancherException as e:
                logger.error('Error blocking user %s. %s' % (user, e))
        return count

    @staticmethod
    def update_users_roles(users):
        count = 0

        for user in users:
            for service_settings in users[user]:
                has_change = False
                rancher_user = models.RancherUser.objects.get(user=user, settings=service_settings)
                current_links = models.RancherUserClusterLink.objects.filter(user=rancher_user)
                actual_links = users[user][service_settings]
                current_links_set = {(link.cluster.id, link.role) for link in current_links}

                actual_links_set = set()

                for link in actual_links:
                    role = models.ClusterRole.CLUSTER_OWNER if link[1] in ['owner', 'manager'] \
                        else models.ClusterRole.CLUSTER_MEMBER if link[1] in ['admin'] else None
                    actual_links_set.add((link[0].id, role))

                remove_links = current_links_set - actual_links_set
                add_links = actual_links_set - current_links_set

                backend = RancherBackend(service_settings)

                for link in remove_links:
                    cluster_id = link[0]
                    role = link[1]
                    rancher_user_cluster_link = models.RancherUserClusterLink.objects.get(
                        user=rancher_user,
                        role=role,
                        cluster_id=cluster_id
                    )

                    try:
                        with transaction.atomic():
                            backend.delete_cluster_role(rancher_user_cluster_link)

                        has_change = True
                    except exceptions.RancherException as e:
                        logger.error('Error deleting role %s. %s' % (rancher_user_cluster_link.id, e))

                for link in add_links:
                    cluster_id = link[0]
                    role = link[1]

                    try:
                        with transaction.atomic():
                            rancher_user_cluster_link = models.RancherUserClusterLink.objects.create(
                                user=rancher_user,
                                role=role,
                                cluster_id=cluster_id
                            )
                            backend.create_cluster_role(rancher_user_cluster_link)

                        has_change = True
                    except exceptions.RancherException as e:
                        logger.error('Error creating role. User ID: %s, cluster ID: %s, role: %s. %s' %
                                     (rancher_user.id, cluster_id, role, e))

                if has_change:
                    count += 1

        return count

    @classmethod
    def run(cls):
        result = {}
        actual_users = cls.get_users()
        current_users = models.RancherUser.objects.filter(is_active=True)
        current_users_set = {(user.user.uuid.hex, user.settings.uuid.hex) for user in current_users}
        actual_users_set = set()

        for user in actual_users:
            for service_settings in actual_users[user]:
                actual_users_set.add((user.uuid.hex, service_settings.uuid.hex))

        # Delete users
        remove_rancher_users_set = current_users_set - actual_users_set
        remove_rancher_users = []

        for user in remove_rancher_users_set:
            user_uuid = user[0]
            settings_uuid = user[1]
            remove_rancher_users.append(
                models.RancherUser.objects.get(user__uuid=user_uuid, settings__uuid=settings_uuid)
            )

        result['blocked'] = cls.block_users(remove_rancher_users)

        # Create users
        add_users_set = actual_users_set - current_users_set
        add_users = {}

        for user in add_users_set:
            user_uuid = user[0]
            settings_uuid = user[1]
            user = User.objects.get(uuid=user_uuid)
            service_settings = ServiceSettings.objects.get(uuid=settings_uuid)

            if user not in add_users.keys():
                add_users[user] = {}

            if service_settings not in add_users[user].keys():
                add_users[user][service_settings] = []

            add_users[user][service_settings].extend(actual_users[user][service_settings])

        result['created'], result['activated'] = cls.create_users(add_users)

        # Update user's roles
        result['updated'] = cls.update_users_roles(actual_users)

        return result



from django.conf import settings
from django.urls import reverse, NoReverseMatch
from django.utils.translation import ugettext_lazy as _
from fluent_dashboard.dashboard import modules, FluentIndexDashboard, FluentAppIndexDashboard

from waldur_core.core import models as core_models
from waldur_core.logging import models as logging_models
from waldur_core.structure import models as structure_models, SupportedServices


class CustomIndexDashboard(FluentIndexDashboard):
    """
    Custom index dashboard for admin site.
    """
    title = _('Waldur administration')

    def _get_quick_access_info(self):
        """
        Returns a list of ListLink items to be added to Quick Access tab.
        Contains:
        - links to Organizations, Projects and Users;
        - a link to shared service settings;
        - custom configured links in admin/settings FLUENT_DASHBOARD_QUICK_ACCESS_LINKS attribute;
        """
        quick_access_links = []

        # add custom links
        quick_access_links.extend(settings.FLUENT_DASHBOARD_QUICK_ACCESS_LINKS)

        for model in (structure_models.Project,
                      structure_models.Customer,
                      core_models.User,
                      structure_models.SharedServiceSettings,
                      logging_models.Report,):
            link = self._get_link_to_model(model)
            if 'url' in link:
                quick_access_links.append(link)

        return quick_access_links

    def _get_erred_resource_link(self, model, erred_amount, erred_state):
        result = self._get_link_to_model(model)
        result['title'] = _('%(num)s %(resources)s in ERRED state') % {
            'num': erred_amount,
            'resources': result['title']
        }
        if 'url' in result:
            result['url'] = '%s?shared__exact=1&state__exact=%s' % (result['url'], erred_state)
        return result

    def _get_link_to_model(self, model):
        result = {
            'title': str(model._meta.verbose_name_plural).capitalize(),
            'external': True,
            'attrs': {'target': '_blank'},
        }
        try:
            result['url'] = reverse('admin:%s_%s_changelist' % (model._meta.app_label, model._meta.model_name))
        except NoReverseMatch:
            pass
        return result

    def _get_link_to_instance(self, instance):
        result = {
            'title': str(instance),
            'external': True,
            'attrs': {'target': '_blank'},
        }
        try:
            result['url'] = reverse('admin:%s_%s_change' % (instance._meta.app_label, instance._meta.model_name),
                                    args=(instance.pk,))
        except NoReverseMatch:
            pass
        return result

    def _get_erred_shared_settings_module(self):
        """
        Returns a LinkList based module which contains link to shared service setting instances in ERRED state.
        """
        result_module = modules.LinkList(title=_('Shared provider settings in erred state'))
        result_module.template = 'admin/dashboard/erred_link_list.html'
        erred_state = structure_models.SharedServiceSettings.States.ERRED

        queryset = structure_models.SharedServiceSettings.objects
        settings_in_erred_state = queryset.filter(state=erred_state).count()

        if settings_in_erred_state:
            result_module.title = '%s (%s)' % (result_module.title, settings_in_erred_state)
            for service_settings in queryset.filter(state=erred_state).iterator():
                module_child = self._get_link_to_instance(service_settings)
                if 'url' in module_child:
                    module_child['error'] = service_settings.error_message
                    result_module.children.append(module_child)
        else:
            result_module.pre_content = _('Nothing found.')

        return result_module

    def _get_erred_resources_module(self):
        """
        Returns a list of links to resources which are in ERRED state and linked to a shared service settings.
        """
        result_module = modules.LinkList(title=_('Resources in erred state'))
        erred_state = structure_models.NewResource.States.ERRED
        children = []

        resource_models = SupportedServices.get_resource_models()
        resources_in_erred_state_overall = 0
        for resource_type, resource_model in resource_models.items():
            queryset = resource_model.objects.filter(service_project_link__service__settings__shared=True)
            erred_amount = queryset.filter(state=erred_state).count()
            if erred_amount:
                resources_in_erred_state_overall = resources_in_erred_state_overall + erred_amount
                link = self._get_erred_resource_link(resource_model, erred_amount, erred_state)
                if 'url' in link:
                    children.append(link)

        if resources_in_erred_state_overall:
            result_module.title = '%s (%s)' % (result_module.title, resources_in_erred_state_overall)
            result_module.children = children
        else:
            result_module.pre_content = _('Nothing found.')

        return result_module

    def __init__(self, **kwargs):
        FluentIndexDashboard.__init__(self, **kwargs)

        self.children.append(modules.LinkList(
            _('Quick access'),
            children=self._get_quick_access_info())
        )

        self.children.append(self._get_erred_shared_settings_module())
        self.children.append(self._get_erred_resources_module())


class CustomAppIndexDashboard(FluentAppIndexDashboard):
    def __init__(self, app_title, models, **kwargs):
        super(CustomAppIndexDashboard, self).__init__(app_title, models, **kwargs)
        path = self._get_app_models_path()
        self.children = [modules.ModelList(title=app_title, models=[path])]

    def _get_app_models_path(self):
        return '%s.models.*' % self.app_title.replace(' ', '.', 1).replace(' ', '_').lower()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-12 09:20
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('users', '0001_squashed_0004'),
    ]

    operations = [
        migrations.AddField(
            model_name='invitation',
            name='approved_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='invitation',
            name='full_name',
            field=models.CharField(blank=True, max_length=100, verbose_name='full name'),
        ),
        migrations.AddField(
            model_name='invitation',
            name='job_title',
            field=models.CharField(blank=True, max_length=40, verbose_name='job title'),
        ),
        migrations.AddField(
            model_name='invitation',
            name='native_name',
            field=models.CharField(blank=True, max_length=100, verbose_name='native name'),
        ),
        migrations.AddField(
            model_name='invitation',
            name='organization',
            field=models.CharField(blank=True, max_length=80, verbose_name='organization'),
        ),
        migrations.AddField(
            model_name='invitation',
            name='phone_number',
            field=models.CharField(blank=True, max_length=255, verbose_name='phone number'),
        ),
        migrations.AddField(
            model_name='invitation',
            name='tax_number',
            field=models.CharField(blank=True, max_length=50, verbose_name='tax number'),
        ),
        migrations.AlterField(
            model_name='invitation',
            name='state',
            field=models.CharField(choices=[('requested', 'Requested'), ('rejected', 'Rejected'), ('pending', 'Pending'), ('accepted', 'Accepted'), ('canceled', 'Canceled'), ('expired', 'Expired')], default='pending', max_length=10),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-01-24 16:24
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_slurm', '0003_allocationusage'),
    ]

    operations = [
        migrations.AlterField(
            model_name='allocation',
            name='cpu_limit',
            field=models.BigIntegerField(default=-1),
        ),
        migrations.AlterField(
            model_name='allocation',
            name='cpu_usage',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='allocation',
            name='gpu_limit',
            field=models.BigIntegerField(default=-1),
        ),
        migrations.AlterField(
            model_name='allocation',
            name='gpu_usage',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='allocation',
            name='ram_limit',
            field=models.BigIntegerField(default=-1),
        ),
        migrations.AlterField(
            model_name='allocation',
            name='ram_usage',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='allocationusage',
            name='cpu_usage',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='allocationusage',
            name='gpu_usage',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='allocationusage',
            name='ram_usage',
            field=models.BigIntegerField(default=0),
        ),
    ]



import copy
import json
from collections import defaultdict

from django import forms
from django.conf import settings
from django.conf.urls import url
from django.contrib import admin, messages
from django.contrib.admin import forms as admin_forms
from django.contrib.admin import widgets
from django.contrib.auth import admin as auth_admin, get_user_model
from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError
from django.forms.utils import flatatt
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.html import format_html_join
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from jsoneditor.forms import JSONEditor
from rest_framework import permissions as rf_permissions
from reversion.admin import VersionAdmin

from waldur_core.core import models
from waldur_core.core.authentication import can_access_admin_site


def get_admin_url(obj):
    return reverse('admin:%s_%s_change' % (obj._meta.app_label, obj._meta.model_name), args=[obj.id])


def render_to_readonly(value):
    return "<p>{0}</p>".format(value)


class ReadonlyTextWidget(forms.TextInput):
    def format_value(self, value):
        return value

    def render(self, name, value, attrs=None, renderer=None):
        return render_to_readonly(self.format_value(value))


class ReadOnlyAdminMixin:
    """
    Disables all editing capabilities.
    Please ensure that readonly_fields is specified in derived class.
    """
    change_form_template = 'admin/core/readonly_change_form.html'

    def get_actions(self, request):
        actions = super(ReadOnlyAdminMixin, self).get_actions(request)
        if 'delete_selected' in actions:
            del actions['delete_selected']
        return actions

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        if request.user.is_staff:
            return True
        return False

    def save_model(self, request, obj, form, change):
        pass

    def delete_model(self, request, obj):
        pass

    def save_related(self, request, form, formsets, change):
        pass


class CopyButtonMixin:
    class Media:
        js = (
            settings.STATIC_URL + 'landing/js/copy2clipboard.js',
        )

    def render(self, name, value, attrs=None, renderer=None):
        result = super().render(name, value, attrs)
        button_attrs = {
            'class': 'button copy-button',
            'data-target-id': attrs['id'],
        }
        result += "<a %(attrs)s>Copy</a>" % {'attrs': flatatt(button_attrs)}
        return mark_safe(result)  # nosec


class PasswordWidget(CopyButtonMixin, forms.PasswordInput):
    template_name = 'admin/core/widgets/password-widget.html'

    def __init__(self, attrs=None):
        super(PasswordWidget, self).__init__(attrs, render_value=True)


class JsonWidget(CopyButtonMixin, JSONEditor):
    class Media:
        js = JSONEditor.Media.js + CopyButtonMixin.Media.js


def format_json_field(value):
    template = '<div><pre style="overflow: hidden">{0}</pre></div>'
    formatted_value = json.dumps(value, indent=True, ensure_ascii=False)
    return mark_safe(template.format(formatted_value))  # nosec


class OptionalChoiceField(forms.ChoiceField):
    def __init__(self, choices=(), *args, **kwargs):
        empty = [('', '---------')]
        choices = empty + sorted(choices, key=lambda pair: pair[1])
        super(OptionalChoiceField, self).__init__(choices=choices, *args, **kwargs)


class UserCreationForm(auth_admin.UserCreationForm):
    class Meta:
        model = get_user_model()
        fields = ("username",)

    # overwritten to support custom User model
    def clean_username(self):
        # Since User.username is unique, this check is redundant,
        # but it sets a nicer error message than the ORM. See #13147.
        username = self.cleaned_data["username"]
        try:
            get_user_model()._default_manager.get(username=username)
        except get_user_model().DoesNotExist:
            return username
        raise forms.ValidationError(
            self.error_messages['duplicate_username'],
            code='duplicate_username',
        )


class UserChangeForm(auth_admin.UserChangeForm):
    class Meta:
        model = get_user_model()
        exclude = ('details',)

    def __init__(self, *args, **kwargs):
        super(UserChangeForm, self).__init__(*args, **kwargs)
        competences = [(key, key) for key in settings.WALDUR_CORE.get('USER_COMPETENCE_LIST', [])]
        self.fields['competence'] = OptionalChoiceField(choices=competences, required=False)

    def clean_civil_number(self):
        # Empty string should be converted to None.
        # Otherwise uniqueness constraint is violated.
        # See also: http://stackoverflow.com/a/1400046/175349
        civil_number = self.cleaned_data.get('civil_number')
        if civil_number:
            return civil_number.strip()
        return None


class ExcludedFieldsAdminMixin(admin.ModelAdmin):
    """
    This mixin allows to toggle display of fields in Django model admin according to custom logic.
    It's expected that inherited class has implemented excluded_fields property.
    """

    @cached_property
    def excluded_fields(self):
        return []

    def filter_excluded_fields(self, fields):
        return [field for field in fields if field not in self.excluded_fields]

    def exclude_fields_from_fieldset(self, fieldset):
        name, options = fieldset
        fields = options.get('fields', ())
        options = copy.copy(options)
        options['fields'] = self.filter_excluded_fields(fields)
        return (name, options)

    def get_fields(self, request, obj=None):
        fields = super(ExcludedFieldsAdminMixin, self).get_fields(request, obj)
        return self.filter_excluded_fields(fields)

    def get_list_display(self, request):
        fields = super(ExcludedFieldsAdminMixin, self).get_list_display(request)
        return self.filter_excluded_fields(fields)

    def get_search_fields(self, request):
        fields = super(ExcludedFieldsAdminMixin, self).get_search_fields(request)
        return self.filter_excluded_fields(fields)

    def get_fieldsets(self, request, obj=None):
        fieldsets = super(ExcludedFieldsAdminMixin, self).get_fieldsets(request, obj)
        return map(self.exclude_fields_from_fieldset, fieldsets)


class NativeNameAdminMixin(ExcludedFieldsAdminMixin):
    @cached_property
    def excluded_fields(self):
        if not settings.WALDUR_CORE['NATIVE_NAME_ENABLED']:
            return ['native_name']
        return []


class UserAdmin(NativeNameAdminMixin, auth_admin.UserAdmin):
    list_display = ('username', 'uuid', 'email', 'full_name', 'native_name', 'is_active', 'is_staff', 'is_support')
    search_fields = ('username', 'uuid', 'full_name', 'native_name', 'email', 'civil_number')
    list_filter = ('is_active', 'is_staff', 'is_support', 'registration_method')
    date_hierarchy = 'date_joined'
    fieldsets = (
        (None, {'fields': ('username', 'password', 'registration_method', 'uuid')}),
        (_('Personal info'), {'fields': (
            'civil_number', 'full_name', 'native_name', 'email',
            'preferred_language', 'competence', 'phone_number'
        )}),
        (_('Organization'), {'fields': ('organization', 'job_title',)}),
        (_('Permissions'), {'fields': ('is_active', 'is_staff', 'is_support', 'customer_roles', 'project_roles')}),
        (_('Important dates'), {'fields': ('last_login', 'date_joined', 'agreement_date')}),
        (_('Authentication backend details'), {'fields': ('format_details', 'backend_id')}),
    )
    readonly_fields = ('registration_method', 'agreement_date', 'customer_roles', 'project_roles', 'uuid',
                       'last_login', 'date_joined', 'format_details')
    form = UserChangeForm
    add_form = UserCreationForm

    def customer_roles(self, instance):
        from waldur_core.structure.models import CustomerPermission
        permissions = CustomerPermission.objects.filter(user=instance, is_active=True).order_by('customer')

        return format_html_join(
            mark_safe('<br/>'),  # nosec
            '<a href={}>{}</a>',
            ((get_admin_url(permission.customer), str(permission)) for permission in permissions),
        ) or mark_safe("<span class='errors'>%s</span>" % _('User has no roles in any organization.'))  # nosec

    customer_roles.short_description = _('Roles in organizations')

    def project_roles(self, instance):
        from waldur_core.structure.models import ProjectPermission
        permissions = ProjectPermission.objects.filter(user=instance, is_active=True).order_by('project')

        return format_html_join(
            mark_safe('<br/>'),  # nosec
            '<a href={}>{}</a>',
            ((get_admin_url(permission.project), str(permission)) for permission in permissions),
        ) or mark_safe("<span class='errors'>%s</span>" % _('User has no roles in any project.'))  # nosec

    project_roles.short_description = _('Roles in projects')

    def format_details(self, obj):
        return format_json_field(obj.details)

    format_details.allow_tags = True
    format_details.short_description = _('Details')


class SshPublicKeyAdmin(admin.ModelAdmin):
    list_display = ('user', 'name', 'fingerprint')
    search_fields = ('user__username', 'name', 'fingerprint')
    readonly_fields = ('user', 'name', 'fingerprint', 'public_key')


class ChangeEmailRequestAdmin(ReadOnlyAdminMixin, admin.ModelAdmin):
    list_display = ('user', 'email', 'created')


class CustomAdminAuthenticationForm(admin_forms.AdminAuthenticationForm):
    error_messages = {
        'invalid_login': _("Please enter the correct %(username)s and password "
                           "for a staff or a support account. Note that both fields may be "
                           "case-sensitive."),
    }

    def confirm_login_allowed(self, user):
        if not can_access_admin_site(user):
            return super(CustomAdminAuthenticationForm, self).confirm_login_allowed(user)


class CustomAdminSite(admin.AdminSite):
    site_title = _('Waldur MasterMind admin')
    site_header = _('Waldur MasterMind administration')
    index_title = _('Waldur MasterMind administration')
    login_form = CustomAdminAuthenticationForm

    def has_permission(self, request):
        is_safe = request.method in rf_permissions.SAFE_METHODS
        return can_access_admin_site(request.user) and (is_safe or request.user.is_staff)

    @classmethod
    def clone_default(cls):
        instance = cls()
        instance._registry = admin.site._registry.copy()
        instance._actions = admin.site._actions.copy()
        instance._global_actions = admin.site._global_actions.copy()
        return instance


admin_site = CustomAdminSite.clone_default()
admin.site = admin_site
admin.site.register(models.User, UserAdmin)
admin.site.register(models.SshPublicKey, SshPublicKeyAdmin)
admin.site.register(models.ChangeEmailRequest, ChangeEmailRequestAdmin)


# TODO: Extract common classes to admin_utils module and remove hack.
# This hack is needed because admin is imported several times.
# Please note that admin module should NOT be imported by other apps.
if admin.site.is_registered(Group):
    admin.site.unregister(Group)


class ReversionAdmin(VersionAdmin):
    def add_view(self, request, form_url='', extra_context=None):
        # Revision creation is ignored in this method because it has to be implemented in model.save method
        return super(VersionAdmin, self).add_view(request, form_url, extra_context)

    def change_view(self, request, object_id, form_url='', extra_context=None):
        # Revision creation is ignored in this method because it has to be implemented in model.save method
        return super(VersionAdmin, self).change_view(request, object_id, form_url, extra_context)


class ExecutorAdminAction:
    """ Add executor as action to admin model.

    Usage example:
        class PullSecurityGroups(ExecutorAdminAction):
            executor = executors.TenantPullSecurityGroupsExecutor  # define executor
            short_description = 'Pull security groups'  # description for admin page
            confirmation = True # if your action requires a confirmation else set False

            def validate(self, tenant):
                if tenant.state != Tenant.States.OK:
                    raise ValidationError('Tenant has to be in state OK to pull security groups.')

        pull_security_groups = PullSecurityGroups()  # this action could be registered as admin action

    """
    executor = NotImplemented
    short_description = ''
    confirmation_template = 'admin/action_confirmation.html'
    confirmation_description = ''
    confirmation = False

    def __call__(self, admin_class, request, queryset):
        if self.confirmation and not request.POST.get('confirmed'):
            return self.confirmation_response(admin_class, request, queryset)
        else:
            return self.execute(admin_class, request, queryset)

    def confirmation_response(self, admin_class, request, queryset):
        opts = admin_class.model._meta
        app_label = opts.app_label
        object_name = str(opts.verbose_name)
        context = {
            **admin_class.admin_site.each_context(request),
            'title': _("Are you sure?"),
            'object_name': object_name,
            'queryset': queryset,
            'opts': opts,
            'app_label': app_label,
            'action_name': self.get_action_name(admin_class),
            'confirmation_description': self.confirmation_description,
            'description': self.short_description,
        }
        request.current_app = admin_class.admin_site.name
        context.update(
            media=admin_class.media,
        )
        return TemplateResponse(request, self.confirmation_template, context, )

    def execute(self, admin_class, request, queryset):
        errors = defaultdict(list)
        successfully_executed = []
        for instance in queryset:
            try:
                self.validate(instance)
            except ValidationError as e:
                errors[str(e)].append(instance)
            else:
                params = self.get_execute_params(request, instance)
                self.executor.execute(instance, **params)
                successfully_executed.append(instance)

        if successfully_executed:
            message = _('Operation was successfully scheduled for %(count)d instances: %(names)s') % dict(
                count=len(successfully_executed),
                names=', '.join([str(i) for i in successfully_executed])
            )
            admin_class.message_user(request, message)

        for error, instances in errors.items():
            message = _('Failed to schedule operation for %(count)d instances: %(names)s. Error: %(message)s') % dict(
                count=len(instances),
                names=', '.join([str(i) for i in instances]),
                message=error,
            )
            admin_class.message_user(request, message, level=messages.ERROR)

    def get_action_name(self, admin_class):

        for action_name in admin_class.actions:
            action_obj = getattr(admin_class, action_name, None)
            if isinstance(action_obj, self.__class__) and action_obj.confirmation:
                return action_name

    def validate(self, instance):
        """ Raise validation error if action cannot be performed for given instance """
        pass

    def get_execute_params(self, request, instance):
        """ Returns additional parameters for the executor """
        return {}


class ExtraActionsMixin:
    """
    Allows to add extra actions to admin list page.
    """
    change_list_template = 'admin/core/change_list.html'

    def get_extra_actions(self):
        raise NotImplementedError('Method "get_extra_actions" should be implemented in ExtraActionsMixin.')

    def get_urls(self):
        """
        Inject extra action URLs.
        """
        urls = []

        for action in self.get_extra_actions():
            regex = r'^{}/$'.format(self._get_action_href(action))
            view = self.admin_site.admin_view(action)
            urls.append(url(regex, view))

        return urls + super(ExtraActionsMixin, self).get_urls()

    def changelist_view(self, request, extra_context=None):
        """
        Inject extra links into template context.
        """
        links = []

        for action in self.get_extra_actions():
            links.append({
                'label': self._get_action_label(action),
                'href': self._get_action_href(action)
            })

        extra_context = extra_context or {}
        extra_context['extra_links'] = links

        return super(ExtraActionsMixin, self).changelist_view(
            request, extra_context=extra_context,
        )

    def _get_action_href(self, action):
        return action.__name__

    def _get_action_label(self, action):
        return getattr(action, 'name', action.__name__.replace('_', ' ').capitalize())


class ExtraActionsObjectMixin:
    """
    Allows to add extra actions to admin object edit page.
    """
    change_form_template = 'admin/core/change_form.html'

    def get_extra_object_actions(self):
        raise NotImplementedError('Method "get_extra_object_actions" should be implemented in ExtraActionsMixin.')

    def get_urls(self):
        """
        Inject extra action URLs.
        """
        urls = []

        for action in self.get_extra_object_actions():
            regex = r'^(.+)/change/{}/$'.format(self._get_action_href(action))
            view = self.admin_site.admin_view(action)
            urls.append(url(regex, view))

        return urls + super(ExtraActionsObjectMixin, self).get_urls()

    def change_view(self, request, object_id, form_url='', extra_context=None):
        """
        Inject extra links into template context.
        """
        links = []
        obj = get_object_or_404(self.model, pk=object_id)

        for action in self.get_extra_object_actions():
            validator = self._get_action_validator(action)
            links.append({
                'label': self._get_action_label(action),
                'href': self._get_action_href(action),
                'show': True if not validator else validator(request, obj)
            })

        extra_context = extra_context or {}
        extra_context['extra_object_links'] = links

        return super(ExtraActionsObjectMixin, self).change_view(
            request, object_id, form_url, extra_context=extra_context,
        )

    def _get_action_href(self, action):
        return action.__name__

    def _get_action_label(self, action):
        return getattr(action, 'name', action.__name__.replace('_', ' ').capitalize())

    def _get_action_validator(self, action):
        return getattr(action, 'validator', None)


class UpdateOnlyModelAdmin:

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        if request.user.is_staff:
            return True
        return False


class GBtoMBWidget(widgets.AdminIntegerFieldWidget):
    def value_from_datadict(self, data, files, name):
        value = super(GBtoMBWidget, self).value_from_datadict(data, files, name) or 0
        value = int(value) * 1024
        return value

    def format_value(self, value):
        return int(value) / 1024

    def render(self, name, value, attrs=None, renderer=None):
        result = super(GBtoMBWidget, self).render(name, value, attrs)
        return '<label>%s GB</label>' % result



from waldur_core.cost_tracking import CostTrackingRegister, CostTrackingStrategy, ConsumableItem

from . import models


class DropletStrategy(CostTrackingStrategy):
    resource_class = models.Droplet

    class Types:
        FLAVOR = 'flavor'

    @classmethod
    def get_consumable_items(cls):
        return [ConsumableItem(item_type=cls.Types.FLAVOR, key=size.name, default_price=size.price)
                for size in models.Size.objects.all()]

    @classmethod
    def get_configuration(cls, droplet):
        consumables = {}
        if droplet.state != models.Droplet.States.ERRED and droplet.size_name:
            consumables[ConsumableItem(item_type=cls.Types.FLAVOR, key=droplet.size_name)] = 1
        return consumables


CostTrackingRegister.register_strategy(DropletStrategy)



from rest_framework import serializers


def validate_options(options, attributes):
    fields = {}

    for name, option in options.items():
        params = {}
        field_type = option.get('type', '')
        field_class = serializers.CharField

        if field_type == 'integer':
            field_class = serializers.IntegerField

        elif field_type == 'money':
            field_class = serializers.IntegerField

        elif field_type == 'boolean':
            field_class = serializers.BooleanField

        default_value = option.get('default')
        if default_value:
            params['default'] = default_value
        else:
            params['required'] = option.get('required', False)

        if field_class == serializers.IntegerField:
            if 'min' in option:
                params['min_value'] = option.get('min')

            if 'max' in option:
                params['max_value'] = option.get('max')

        if 'choices' in option:
            field_class = serializers.ChoiceField
            params['choices'] = option.get('choices')

        fields[name] = field_class(**params)

    serializer_class = type('AttributesSerializer', (serializers.Serializer,), fields)
    serializer = serializer_class(data=attributes)
    serializer.is_valid(raise_exception=True)



# Generated by Django 2.2.6 on 2019-10-23 09:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0019_immutable_default_json'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attachment',
            name='file',
            field=models.FileField(upload_to='jira_attachments'),
        ),
    ]



import functools
import logging
from datetime import timedelta

from celery import shared_task
from django.conf import settings
from django.db import transaction
from django.db.utils import DatabaseError
from django.utils import timezone

from waldur_core.core import utils as core_utils, tasks as core_tasks, models as core_models
from waldur_core.quotas.exceptions import QuotaValidationError
from waldur_core.structure import SupportedServices, models, ServiceBackendError, models as structure_models

logger = logging.getLogger(__name__)


def reraise_exceptions(func):
    @functools.wraps(func)
    def wrapped(self, service_settings, *args, **kwargs):
        try:
            return func(self, service_settings, *args, **kwargs)
        except Exception as e:
            raise e.__class__(
                '%s, Service settings: %s, %s' % (e, service_settings.name, service_settings.type)
            )

    return wrapped


@shared_task(name='waldur_core.structure.check_expired_permissions')
def check_expired_permissions():
    for cls in models.BasePermission.get_all_models():
        for permission in cls.get_expired():
            permission.revoke()


def connect_shared_settings(service_settings):
    logger.debug('About to connect service settings "%s" to all available customers' % service_settings.name)
    if not service_settings.shared:
        raise ValueError('It is impossible to connect non-shared settings')
    service_model = SupportedServices.get_service_models()[service_settings.type]['service']

    with transaction.atomic():
        for customer in models.Customer.objects.all():
            defaults = {'available_for_all': True}
            try:
                service, _ = service_model.objects.get_or_create(
                    customer=customer, settings=service_settings, defaults=defaults)
            except QuotaValidationError:
                logger.warning('Unable to connect shared service '
                               'settings to customer because quota is exceeded. '
                               'Service settings ID: %s, customer ID: %s',
                               service_settings.id, customer.id)
                continue

            service_project_link_model = service.projects.through
            for project in service.customer.projects.all():
                try:
                    service_project_link_model.objects.get_or_create(project=project, service=service)
                except QuotaValidationError:
                    logger.warning('Unable to connect shared service to project because '
                                   'quota is exceeded. Service ID: %s, project ID: %s',
                                   service.id, project.id)
                    continue


class ConnectSharedSettingsTask(core_tasks.Task):

    def execute(self, service_settings):
        connect_shared_settings(service_settings)
        logger.info('Successfully connected service settings "%s" to all available customers' % service_settings.name)


class BackgroundPullTask(core_tasks.BackgroundTask):
    """ Pull information about object from backend. Method "pull" should be implemented.

        Task marks object as ERRED if pull failed and recovers it if pull succeed.
    """

    def run(self, serialized_instance):
        instance = core_utils.deserialize_instance(serialized_instance)
        try:
            self.pull(instance)
        except ServiceBackendError as e:
            self.on_pull_fail(instance, e)
        else:
            self.on_pull_success(instance)

    def is_equal(self, other_task, serialized_instance):
        return self.name == other_task.get('name') and serialized_instance in other_task.get('args', [])

    def pull(self, instance):
        """ Pull instance from backend.

            This method should not handle backend exception.
        """
        raise NotImplementedError('Pull task should implement pull method.')

    def on_pull_fail(self, instance, error):
        error_message = str(error)
        self.log_error_message(instance, error_message)
        try:
            self.set_instance_erred(instance, error_message)
        except DatabaseError as e:
            logger.debug(e, exc_info=True)

    def on_pull_success(self, instance):
        if instance.state == instance.States.ERRED:
            instance.recover()
            instance.error_message = ''
            instance.save(update_fields=['state', 'error_message'])

    def log_error_message(self, instance, error_message):
        logger_message = 'Failed to pull %s %s (PK: %s). Error: %s' % (
            instance.__class__.__name__, instance.name, instance.pk, error_message)
        if instance.state == instance.States.ERRED:  # report error on debug level if instance already was erred.
            logger.debug(logger_message)
        else:
            logger.error(logger_message, exc_info=True)

    def set_instance_erred(self, instance, error_message):
        """ Mark instance as erred and save error message """
        instance.set_erred()
        instance.error_message = error_message
        instance.save(update_fields=['state', 'error_message'])


class BackgroundListPullTask(core_tasks.BackgroundTask):
    """ Schedules pull task for each stable object of the model. """
    model = NotImplemented
    pull_task = NotImplemented

    def is_equal(self, other_task):
        return self.name == other_task.get('name')

    def get_pulled_objects(self):
        States = self.model.States
        return self.model.objects.filter(state__in=[States.ERRED, States.OK]).exclude(backend_id='')

    def run(self):
        for instance in self.get_pulled_objects():
            serialized = core_utils.serialize_instance(instance)
            self.pull_task().apply_async(args=(serialized,), kwargs={})


class ServiceListPullTask(BackgroundListPullTask):
    model = models.ServiceSettings

    def get_pulled_objects(self):
        States = self.model.States
        return self.model.objects.filter(state__in=[States.ERRED, States.OK])


class ServicePropertiesPullTask(BackgroundPullTask):

    def pull(self, service_settings):
        backend = service_settings.get_backend()
        backend.pull_service_properties()


class ServiceResourcesPullTask(BackgroundPullTask):

    @reraise_exceptions
    def pull(self, service_settings):
        backend = service_settings.get_backend()
        backend.pull_resources()


class ServiceSubResourcesPullTask(BackgroundPullTask):

    def pull(self, service_settings):
        backend = service_settings.get_backend()
        backend.pull_subresources()


class ServicePropertiesListPullTask(ServiceListPullTask):
    name = 'waldur_core.structure.ServicePropertiesListPullTask'
    pull_task = ServicePropertiesPullTask


class ServiceResourcesListPullTask(ServiceListPullTask):
    name = 'waldur_core.structure.ServiceResourcesListPullTask'
    pull_task = ServiceResourcesPullTask


class ServiceSubResourcesListPullTask(ServiceListPullTask):
    name = 'waldur_core.structure.ServiceSubResourcesListPullTask'
    pull_task = ServiceSubResourcesPullTask


class RetryUntilAvailableTask(core_tasks.Task):
    max_retries = 300
    default_retry_delay = 5

    def pre_execute(self, instance):
        if not self.is_available(instance):
            self.retry()
        super(RetryUntilAvailableTask, self).pre_execute(instance)

    def is_available(self, instance):
        return True


class BaseThrottleProvisionTask(RetryUntilAvailableTask):
    """
    Before starting resource provisioning, count how many resources
    are already in "creating" state and delay provisioning if there are too many of them.
    """
    DEFAULT_LIMIT = 4

    def is_available(self, resource):
        usage = self.get_usage(resource)
        limit = self.get_limit(resource)
        return usage <= limit

    def get_usage(self, resource):
        service_settings = resource.service_project_link.service.settings
        model_class = resource._meta.model
        return model_class.objects.filter(
            state=core_models.StateMixin.States.CREATING,
            service_project_link__service__settings=service_settings).count()

    def get_limit(self, resource):
        return self.DEFAULT_LIMIT


class ThrottleProvisionTask(BaseThrottleProvisionTask, core_tasks.BackendMethodTask):
    pass


class ThrottleProvisionStateTask(BaseThrottleProvisionTask, core_tasks.StateTransitionTask):
    pass


class SetErredStuckResources(core_tasks.BackgroundTask):
    """
    This task marks all resources which have been provisioning for more than 3 hours as erred.
    """
    name = 'waldur_core.structure.SetErredStuckResources'

    def is_equal(self, other_task):
        return self.name == other_task.get('name')

    def run(self):
        cutoff = timezone.now() - timedelta(hours=3)
        states = (structure_models.NewResource.States.CREATING,
                  structure_models.NewResource.States.CREATION_SCHEDULED)
        resource_models = (structure_models.NewResource.get_all_models() +
                           structure_models.SubResource.get_all_models())
        for model in resource_models:
            for resource in model.objects.filter(modified__lt=cutoff, state__in=states):
                resource.set_erred()
                resource.error_message = 'Provisioning has timed out.'
                resource.save(update_fields=['state', 'error_message'])
                logger.warning('Switching resource %s to erred state, '
                               'because provisioning has timed out.',
                               core_utils.serialize_instance(resource))


@shared_task
def send_change_email_notification(request_serialized):
    request = core_utils.deserialize_instance(request_serialized)
    link = settings.WALDUR_CORE['EMAIL_CHANGE_URL'].format(code=request.uuid.hex)
    context = {'request': request, 'link': link}
    core_utils.broadcast_mail('structure', 'change_email_request', context, [request.email])



import logging

from waldur_core.core import utils as core_utils

from . import tasks, mixins

logger = logging.getLogger(__name__)


def create_doi(instance):
    if isinstance(instance, mixins.DataciteMixin):
        serialized_instance = core_utils.serialize_instance(instance)
        tasks.create_doi.delay(serialized_instance)
    else:
        logger.warning('Instance %s is not DataciteMixin item.' % instance)



import logging

from django.apps import apps
from django.db import models
from django.utils.lru_cache import lru_cache

logger = logging.getLogger(__name__)


class CoordinatesMixin(models.Model):
    """
    Mixin to add a latitude and longitude fields
    """
    class Meta:
        abstract = True

    latitude = models.FloatField(null=True, blank=True)
    longitude = models.FloatField(null=True, blank=True)


class IPCoordinatesMixin(CoordinatesMixin):
    detect_coordinates = NotImplemented

    class Meta:
        abstract = True

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]



import json

import django_filters
from django.utils.translation import ugettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from django_filters.widgets import BooleanWidget
from rest_framework import exceptions as rf_exceptions
from rest_framework.filters import BaseFilterBackend

from waldur_core.core import filters as core_filters
from waldur_core.core.utils import is_uuid_like
from waldur_core.structure import filters as structure_filters
from waldur_core.structure import models as structure_models

from . import models


class ServiceProviderFilter(django_filters.FilterSet):
    customer = core_filters.URLFilter(view_name='customer-detail', field_name='customer__uuid')
    customer_uuid = django_filters.UUIDFilter(field_name='customer__uuid')
    o = django_filters.OrderingFilter(fields=(('customer__name', 'customer_name'),))

    class Meta:
        model = models.ServiceProvider
        fields = []


class OfferingFilter(django_filters.FilterSet):
    name = django_filters.CharFilter(lookup_expr='icontains')
    name_exact = django_filters.CharFilter(field_name='name')
    customer = core_filters.URLFilter(view_name='customer-detail', field_name='customer__uuid')
    customer_uuid = django_filters.UUIDFilter(field_name='customer__uuid')
    project_uuid = django_filters.UUIDFilter(method='filter_project')
    allowed_customer_uuid = django_filters.UUIDFilter(field_name='customer__uuid', method='filter_allowed_customer')
    attributes = django_filters.CharFilter(field_name='attributes', method='filter_attributes')
    state = core_filters.MappedMultipleChoiceFilter(
        choices=[(representation, representation) for db_value, representation in models.Offering.States.CHOICES],
        choice_mappings={representation: db_value for db_value, representation in models.Offering.States.CHOICES},
    )
    category_uuid = django_filters.UUIDFilter(field_name='category__uuid')
    billable = django_filters.BooleanFilter(widget=BooleanWidget)
    shared = django_filters.BooleanFilter(widget=BooleanWidget)
    o = django_filters.OrderingFilter(fields=('name', 'created'))

    def filter_allowed_customer(self, queryset, name, value):
        return queryset.filter_for_customer(value)

    def filter_project(self, queryset, name, value):
        return queryset.filter_for_project(value)

    def filter_attributes(self, queryset, name, value):
        try:
            value = json.loads(value)
        except ValueError:
            raise rf_exceptions.ValidationError(_('Filter attribute is not valid json.'))

        if not isinstance(value, dict):
            raise rf_exceptions.ValidationError(_('Filter attribute should be an dict.'))

        for k, v in value.items():
            if isinstance(v, list):
                # If a filter value is a list, use multiple choice.
                queryset = queryset.filter(**{'attributes__{key}__has_any_keys'.format(key=k): v})
            else:
                queryset = queryset.filter(attributes__contains={k: v})
        return queryset

    class Meta:
        model = models.Offering
        fields = ['shared', 'type']


class OfferingCustomersFilterBackend(DjangoFilterBackend):
    def filter_queryset(self, request, queryset, view):
        return queryset.filter_for_user(request.user)


class OfferingImportableFilterBackend(DjangoFilterBackend):
    def filter_queryset(self, request, queryset, view):
        if 'importable' in request.query_params:
            return queryset.filter_importable(request.user)
        return queryset


class ScreenshotFilter(django_filters.FilterSet):
    offering = core_filters.URLFilter(view_name='marketplace-offering-detail', field_name='offering__uuid')
    offering_uuid = django_filters.UUIDFilter(field_name='offering__uuid')

    o = django_filters.OrderingFilter(fields=('name', 'created'))

    class Meta:
        model = models.Screenshot
        fields = []


class CartItemFilter(django_filters.FilterSet):
    customer = core_filters.URLFilter(view_name='customer-detail', field_name='project__customer__uuid')
    customer_uuid = django_filters.UUIDFilter(field_name='project__customer__uuid')
    project = core_filters.URLFilter(view_name='project-detail', field_name='project__uuid')
    project_uuid = django_filters.UUIDFilter(field_name='project__uuid')

    class Meta:
        model = models.CartItem
        fields = []


class OrderFilter(django_filters.FilterSet):
    customer = core_filters.URLFilter(view_name='customer-detail', field_name='project__customer__uuid')
    customer_uuid = django_filters.UUIDFilter(field_name='project__customer__uuid')
    project = core_filters.URLFilter(view_name='project-detail', field_name='project__uuid')
    project_uuid = django_filters.UUIDFilter(field_name='project__uuid')
    state = core_filters.MappedMultipleChoiceFilter(
        choices=[(representation, representation) for db_value, representation in models.Order.States.CHOICES],
        choice_mappings={representation: db_value for db_value, representation in models.Order.States.CHOICES},
    )
    o = django_filters.OrderingFilter(fields=('created', 'approved_at', 'total_cost', 'state'))

    class Meta:
        model = models.Order
        fields = []


class OrderItemFilter(django_filters.FilterSet):
    offering = core_filters.URLFilter(view_name='marketplace-offering-detail', field_name='offering__uuid')
    offering_uuid = django_filters.UUIDFilter(field_name='offering__uuid')
    project_uuid = django_filters.UUIDFilter(field_name='order__project__uuid')
    category_uuid = django_filters.UUIDFilter(field_name='offering__category__uuid')
    provider_uuid = django_filters.UUIDFilter(field_name='offering__customer__uuid')
    customer_uuid = django_filters.UUIDFilter(field_name='order__project__customer__uuid')
    state = core_filters.MappedMultipleChoiceFilter(
        choices=[(representation, representation) for db_value, representation in models.OrderItem.States.CHOICES],
        choice_mappings={representation: db_value for db_value, representation in models.OrderItem.States.CHOICES},
    )
    type = core_filters.MappedMultipleChoiceFilter(
        choices=[(representation, representation) for db_value, representation in models.OrderItem.Types.CHOICES],
        choice_mappings={representation: db_value for db_value, representation in models.OrderItem.Types.CHOICES},
    )
    order = core_filters.URLFilter(view_name='marketplace-order-detail', field_name='order__uuid')
    order_uuid = django_filters.UUIDFilter(field_name='order__uuid')

    resource = core_filters.URLFilter(view_name='marketplace-resource-detail', field_name='resource__uuid')
    resource_uuid = django_filters.UUIDFilter(field_name='resource__uuid')

    o = django_filters.OrderingFilter(fields=('created',))

    class Meta:
        model = models.OrderItem
        fields = []


class ResourceFilter(django_filters.FilterSet):
    name = django_filters.CharFilter(lookup_expr='icontains')
    name_exact = django_filters.CharFilter(field_name='name')
    query = django_filters.CharFilter(method='filter_query')
    offering = core_filters.URLFilter(view_name='marketplace-offering-detail', field_name='offering__uuid')
    offering_uuid = django_filters.UUIDFilter(field_name='offering__uuid')
    offering_type = django_filters.CharFilter(field_name='offering__type')
    offering_billable = django_filters.UUIDFilter(field_name='offering__billable')
    project_uuid = django_filters.UUIDFilter(field_name='project__uuid')
    project_name = django_filters.CharFilter(field_name='project__name')
    customer_uuid = django_filters.UUIDFilter(field_name='project__customer__uuid')
    category_uuid = django_filters.UUIDFilter(field_name='offering__category__uuid')
    provider_uuid = django_filters.UUIDFilter(field_name='offering__customer__uuid')
    state = core_filters.MappedMultipleChoiceFilter(
        choices=[(representation, representation) for db_value, representation in models.Resource.States.CHOICES],
        choice_mappings={representation: db_value for db_value, representation in models.Resource.States.CHOICES},
    )
    o = django_filters.OrderingFilter(fields=('name', 'created',))

    class Meta:
        model = models.Resource
        fields = []

    def filter_query(self, queryset, name, value):
        if is_uuid_like(value):
            return queryset.filter(uuid=value)
        else:
            return queryset.filter(name__icontains=value)


class ResourceScopeFilterBackend(core_filters.GenericKeyFilterBackend):

    def get_related_models(self):
        return []

    def get_field_name(self):
        return 'scope'


class PlanFilter(django_filters.FilterSet):
    offering = core_filters.URLFilter(view_name='marketplace-offering-detail', field_name='offering__uuid')
    offering_uuid = django_filters.UUIDFilter(field_name='offering__uuid')

    class Meta:
        model = models.Plan
        fields = []


class CategoryComponentUsageScopeFilterBackend(core_filters.GenericKeyFilterBackend):

    def get_related_models(self):
        return [structure_models.Project, structure_models.Customer]

    def get_field_name(self):
        return 'scope'


class CategoryComponentUsageFilter(django_filters.FilterSet):
    class Meta:
        model = models.CategoryComponentUsage
        fields = []

    date_before = django_filters.DateFilter(field_name='date', lookup_expr='lte')
    date_after = django_filters.DateFilter(field_name='date', lookup_expr='gte')


class ComponentUsageFilter(django_filters.FilterSet):
    resource = core_filters.URLFilter(view_name='marketplace-resource-detail', field_name='resource__uuid')
    resource_uuid = django_filters.UUIDFilter(field_name='resource__uuid')
    offering_uuid = django_filters.UUIDFilter(field_name='resource__offering__uuid')
    project_uuid = django_filters.UUIDFilter(field_name='resource__project__uuid')
    customer_uuid = django_filters.UUIDFilter(field_name='resource__project__customer__uuid')
    date_before = django_filters.DateFilter(field_name='date', lookup_expr='lte')
    date_after = django_filters.DateFilter(field_name='date', lookup_expr='gte')
    type = django_filters.CharFilter(field_name='component__type')

    class Meta:
        model = models.ComponentUsage
        fields = []


class OfferingFileFilter(django_filters.FilterSet):
    offering = core_filters.URLFilter(view_name='marketplace-offering-detail', field_name='offering__uuid')
    offering_uuid = django_filters.UUIDFilter(field_name='offering__uuid')

    o = django_filters.OrderingFilter(fields=('name', 'created'))

    class Meta:
        model = models.OfferingFile
        fields = []


class ExternalOfferingFilterBackend(core_filters.ExternalFilterBackend):
    pass


class CustomerResourceFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        if 'has_resources' in request.query_params:
            customers = models.Resource.objects.all().values_list('project__customer_id', flat=True)
            queryset = queryset.filter(pk__in=customers)
        return queryset


class ServiceProviderOfferingFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        customer_uuid = request.query_params.get('service_provider_uuid')

        if customer_uuid and is_uuid_like(customer_uuid):
            customers = models.Resource.objects.filter(offering__customer__uuid=customer_uuid)\
                .values_list('project__customer_id', flat=True)
            queryset = queryset.filter(pk__in=customers)
        return queryset


structure_filters.ExternalCustomerFilterBackend.register(CustomerResourceFilter())
structure_filters.ExternalCustomerFilterBackend.register(ServiceProviderOfferingFilter())



import json

from decimal import Decimal, ROUND_UP
from urllib.parse import urlencode

from dateutil import parser
from django.utils.timezone import get_current_timezone
from rest_framework.test import APIRequestFactory

from waldur_core.core.views import RefreshTokenMixin


def quantize_price(value):
    """
    Returns value rounded up to 2 places after the decimal point.
    :rtype: Decimal
    """
    return value.quantize(Decimal('0.01'), rounding=ROUND_UP)


def get_headers(user):
    """
    It is assumed that localhost is specified in ALLOWED_HOSTS Django setting
    so that internal API requests are allowed.
    """
    token = RefreshTokenMixin().refresh_token(user)
    return dict(
        content_type='application/json',
        HTTP_AUTHORIZATION='Token %s' % token.key,
        SERVER_NAME='localhost'
    )


def get_request(view, user, **extra):
    factory = APIRequestFactory()
    request = factory.get('/', **get_headers(user))
    return view(request, **extra)


def create_request(view, user, post_data, **kwargs):
    factory = APIRequestFactory()
    request = factory.post('/', data=json.dumps(post_data), **get_headers(user))
    return view(request, **kwargs)


def delete_request(view, user, query_params='', **extra):
    factory = APIRequestFactory()
    path = ''
    if query_params:
        path = '?' + urlencode(query_params)
    request = factory.delete(path, **get_headers(user))
    return view(request, **extra)


def parse_datetime(timestr):
    return parser.parse(timestr).replace(tzinfo=get_current_timezone())


def parse_date(timestr):
    return parse_datetime(timestr).date()


def mb_to_gb(value):
    # In marketplace RAM and storage is stored in GB, but in plugin it is stored in MB.
    return quantize_price(Decimal(value / 1024.0))



import django_filters

from waldur_core.core import filters as core_filters
from waldur_core.structure import filters as structure_filters

from . import models


class ServiceProjectLinkFilter(structure_filters.BaseServiceProjectLinkFilter):
    service = core_filters.URLFilter(view_name='rancher-detail', field_name='service__uuid')

    class Meta(structure_filters.BaseServiceProjectLinkFilter.Meta):
        model = models.RancherServiceProjectLink


class ClusterFilter(structure_filters.BaseResourceFilter):
    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Cluster


class NodeFilter(django_filters.FilterSet):
    cluster_uuid = django_filters.UUIDFilter(field_name='cluster__uuid')

    class Meta:
        model = models.Node
        fields = ('cluster_uuid',)



import logging

from celery import current_task
from dateutil.relativedelta import relativedelta
from django.utils import timezone

from waldur_core.core import utils as core_utils
from waldur_core.cost_tracking import models, CostTrackingRegister, ResourceNotRegisteredError
from waldur_core.structure import models as structure_models

logger = logging.getLogger(__name__)


def scope_deletion(sender, instance, **kwargs):
    """ Run different actions on price estimate scope deletion.

        If scope is a customer - delete all customer estimates and their children.
        If scope is a deleted resource - redefine consumption details, recalculate
                                         ancestors estimates and update estimate details.
        If scope is a unlinked resource - delete all resource price estimates and update ancestors.
        In all other cases - update price estimate details.
    """

    is_resource = isinstance(instance, structure_models.ResourceMixin)
    if is_resource and getattr(instance, 'PERFORM_UNLINK', False):
        _resource_unlink(resource=instance)
    elif is_resource and not getattr(instance, 'PERFORM_UNLINK', False):
        _resource_deletion(resource=instance)
    elif isinstance(instance, structure_models.Customer):
        _customer_deletion(customer=instance)
    else:
        for price_estimate in models.PriceEstimate.objects.filter(scope=instance):
            price_estimate.init_details()


def _resource_unlink(resource):
    if resource.__class__ not in CostTrackingRegister.registered_resources:
        return
    for price_estimate in models.PriceEstimate.objects.filter(scope=resource):
        price_estimate.update_ancestors_total(diff=-price_estimate.total)
        price_estimate.delete()


def _customer_deletion(customer):
    for estimate in models.PriceEstimate.objects.filter(scope=customer):
        for descendant in estimate.get_descendants():
            descendant.delete()


def _resource_deletion(resource):
    """ Recalculate consumption details and save resource details """
    if resource.__class__ not in CostTrackingRegister.registered_resources:
        return
    new_configuration = {}
    price_estimate = models.PriceEstimate.update_resource_estimate(resource, new_configuration)
    price_estimate.init_details()


def _is_in_celery_task():
    """ Return True if current code is executed in celery task """
    return bool(current_task)


def resource_update(sender, instance, created=False, **kwargs):
    """ Update resource consumption details and price estimate if its configuration has changed.
        Create estimates for previous months if resource was created not in current month.
    """
    resource = instance
    try:
        new_configuration = CostTrackingRegister.get_configuration(resource)
    except ResourceNotRegisteredError:
        return
    models.PriceEstimate.update_resource_estimate(
        resource, new_configuration, raise_exception=not _is_in_celery_task())
    # Try to create historical price estimates
    if created:
        _create_historical_estimates(resource, new_configuration)


def resource_quota_update(sender, instance, **kwargs):
    """ Update resource consumption details and price estimate if its configuration has changed """
    quota = instance
    resource = quota.scope
    try:
        new_configuration = CostTrackingRegister.get_configuration(resource)
    except ResourceNotRegisteredError:
        return
    models.PriceEstimate.update_resource_estimate(
        resource, new_configuration, raise_exception=not _is_in_celery_task())


def _create_historical_estimates(resource, configuration):
    """ Create consumption details and price estimates for past months.

        Usually we need to update historical values on resource import.
    """
    today = timezone.now()
    month_start = core_utils.month_start(today)
    while month_start > resource.created:
        month_start -= relativedelta(months=1)
        models.PriceEstimate.create_historical(resource, configuration, max(month_start, resource.created))



from waldur_core.logging.loggers import EventLogger, event_logger

from waldur_mastermind.marketplace.models import Resource


class BookingEventLogger(EventLogger):
    resource = Resource

    class Meta:
        event_types = ('device_booking_is_accepted',)

    @staticmethod
    def get_scopes(event_context):
        resource = event_context['resource']
        return {resource, resource.project, resource.project.customer}


event_logger.register('waldur_booking', BookingEventLogger)



import argparse
from csv import DictReader

from django.core.management.base import BaseCommand, CommandError

from ... import models


class Command(BaseCommand):
    help_text = "Import catalog of Amazon images."

    def add_arguments(self, parser):
        parser.add_argument(
            'file',
            type=argparse.FileType('r'),
            metavar='FILE',
            help='AMI catalog file.'
        )
        parser.add_argument(
            '-y', '--yes',
            action='store_true',
            dest='yes',
            default=False,
            help='The answer to any question which would be asked will be yes.'
        )

    def handle(self, *args, **options):
        with open(options['file']) as csvfile:
            data = list(DictReader(csvfile))

        csv_regions = set([image['region'] for image in data])
        nc_regions = {region.name: region.id for region in models.Region.objects.all()}
        new_regions = csv_regions - set(nc_regions.keys())
        if new_regions:
            raise CommandError('%s regions are missing in the database.' % ', '.join(new_regions))

        csv_images = {image['backend_id']: image for image in data}
        csv_ids = set(csv_images.keys())

        nc_images = {image.backend_id: image for image in models.Image.objects.all()}
        nc_ids = set(nc_images.keys())

        new_ids = csv_ids - nc_ids
        if new_ids:
            new_ids_list = ', '.join(sorted(new_ids))
            self.stdout.write('The following AMIs would be created: {}.'.format(new_ids_list))

        common_ids = nc_ids & csv_ids
        updated_ids = set()
        for image_id in common_ids:
            csv_image = csv_images[image_id]
            nc_image = nc_images[image_id]
            csv_region = nc_regions.get(csv_image['region'])
            if nc_image.name != csv_image['name'] or nc_image.region_id != csv_region:
                updated_ids.add(image_id)

        if updated_ids:
            updated_ids_list = ', '.join(sorted(updated_ids))
            self.stdout.write('The following AMIs would be updated: {}'.format(updated_ids_list))

        stale_ids = nc_ids - csv_ids
        if stale_ids:
            stale_ids_list = ', '.join(sorted(stale_ids))
            self.stdout.write('The following AMIs would be deleted: {}'.format(stale_ids_list))

        if not new_ids and not stale_ids and not updated_ids:
            self.stdout.write('There are no changes to apply.')
            return

        if not options['yes']:
            confirm = input('Enter [y] to continue: ')
            if confirm.strip().lower() != 'y':
                self.stdout.write('Changes are not applied.')
                return

        for image_id in new_ids:
            csv_image = csv_images[image_id]
            models.Image.objects.create(name=csv_image['name'],
                                        backend_id=csv_image['backend_id'],
                                        region_id=nc_regions.get(csv_image['region']))

        for image_id in updated_ids:
            csv_image = csv_images[image_id]
            models.Image.objects.filter(backend_id=image_id).update(
                name=csv_image['name'],
                region_id=nc_regions.get(csv_image['region']))

        models.Image.objects.filter(backend_id__in=stale_ids).delete()
        self.stdout.write('All changes are applied.')



import logging

import requests

from waldur_vmware.exceptions import VMwareError

logger = logging.getLogger(__name__)


class VMwareClient:
    """
    Lightweight VMware vCenter Automation API client.
    See also: https://vmware.github.io/vsphere-automation-sdk-rest/vsphere/
    """

    def __init__(self, host, verify_ssl=True):
        """
        Initialize client with connection options.

        :param host: VMware vCenter server IP address / FQDN
        :type host: string
        :param verify_ssl: verify SSL certificates for HTTPS requests
        :type verify_ssl: bool
        """
        self._host = host
        self._base_url = 'https://{0}/rest'.format(self._host)
        self._session = requests.Session()
        self._session.verify = verify_ssl

    def _request(self, method, endpoint, json=None, **kwargs):
        url = '%s/%s' % (self._base_url, endpoint)
        if json:
            json = {'spec': json}
        try:
            response = self._session.request(method, url, json=json, **kwargs)
        except requests.RequestException as e:
            raise VMwareError(e)

        status_code = response.status_code
        if status_code in (requests.codes.ok,
                           requests.codes.created,
                           requests.codes.accepted,
                           requests.codes.no_content):
            if response.content:
                data = response.json()
                if isinstance(data, dict) and 'value' in data:
                    return data['value']
                return data
        else:
            raise VMwareError(response.content)

    def _get(self, endpoint, **kwargs):
        return self._request('get', endpoint, **kwargs)

    def _post(self, endpoint, **kwargs):
        return self._request('post', endpoint, **kwargs)

    def _patch(self, endpoint, **kwargs):
        return self._request('patch', endpoint, **kwargs)

    def _delete(self, endpoint, **kwargs):
        return self._request('delete', endpoint, **kwargs)

    def login(self, username, password):
        """
        Login to vCenter server using username and password.

        :param username: user to connect
        :type username: string
        :param password: password of the user
        :type password: string
        :raises Unauthorized: raised if credentials are invalid.
        """
        self._post('com/vmware/cis/session', auth=(username, password))
        logger.info('Successfully logged in as {0}'.format(username))

    def list_clusters(self):
        return self._get('vcenter/cluster')

    def get_cluster(self, cluster_id):
        return self._get('vcenter/cluster/{0}'.format(cluster_id))

    def list_datacenters(self):
        return self._get('vcenter/datacenter')

    def list_datastores(self):
        return self._get('vcenter/datastore')

    def list_resource_pools(self):
        return self._get('vcenter/resource-pool')

    def list_networks(self):
        return self._get('vcenter/network')

    def list_folders(self, folder_type=None):
        """
        Returns information about folders in vCenter.
        :param folder_type: Type (DATACENTER, DATASTORE, HOST, NETWORK, VIRTUAL_MACHINE) of the vCenter Server folder.
        :rtype: List[Dict]
        """
        params = {}
        if folder_type:
            params['filter.type'] = folder_type
        return self._get('vcenter/folder', params=params)

    def list_vms(self):
        """
        Get all the VMs from vCenter inventory.
        """
        return self._get('vcenter/vm')

    def get_vm(self, vm_id):
        """
        Returns information about a virtual machine.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        """
        return self._get('vcenter/vm/{}'.format(vm_id))

    def create_vm(self, spec):
        """
        Creates a virtual machine.

        :param spec: new virtual machine specification
        :type spec: dict
        :return: Virtual machine identifier
        :rtype: string
        """
        return self._post('vcenter/vm', json=spec)

    def delete_vm(self, vm_id):
        """
        Deletes a virtual machine.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        """
        return self._delete('vcenter/vm/{}'.format(vm_id))

    def start_vm(self, vm_id):
        """
        Power on given virtual machine.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        """
        return self._post('vcenter/vm/{}/power/start'.format(vm_id))

    def stop_vm(self, vm_id):
        """
        Power off given virtual machine.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        """
        return self._post('vcenter/vm/{}/power/stop'.format(vm_id))

    def reset_vm(self, vm_id):
        """
        Resets a powered-on virtual machine.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        """
        return self._post('vcenter/vm/{}/power/reset'.format(vm_id))

    def suspend_vm(self, vm_id):
        """
        Suspends a powered-on virtual machine.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        """
        return self._post('vcenter/vm/{}/power/suspend'.format(vm_id))

    def get_guest_power(self, vm_id):
        """
        Returns information about the guest operating system power state.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        """
        return self._get('vcenter/vm/{}/guest/power'.format(vm_id))

    def shutdown_guest(self, vm_id):
        """
        Issues a request to the guest operating system asking
        it to perform a clean shutdown of all services.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        """
        return self._post('vcenter/vm/{}/guest/power?action=shutdown'.format(vm_id))

    def reboot_guest(self, vm_id):
        """
        Issues a request to the guest operating system asking it to perform a reboot.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        """
        return self._post('vcenter/vm/{}/guest/power?action=reboot'.format(vm_id))

    def get_cpu(self, vm_id):
        """
        Returns the CPU-related settings of a virtual machine.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        """
        return self._get('vcenter/vm/{}/hardware/cpu'.format(vm_id))

    def update_cpu(self, vm_id, spec):
        """
        Updates the CPU-related settings of a virtual machine.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        :param spec: CPU specification
        :type spec: dict
        """
        return self._patch('vcenter/vm/{}/hardware/cpu'.format(vm_id), json=spec)

    def get_memory(self, vm_id):
        """
        Returns the memory-related settings of a virtual machine.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        """
        return self._get('vcenter/vm/{}/hardware/memory'.format(vm_id))

    def update_memory(self, vm_id, spec):
        """
        Updates the memory-related settings of a virtual machine.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        :param spec: CPU specification
        :type spec: dict
        """
        return self._patch('vcenter/vm/{}/hardware/memory'.format(vm_id), json=spec)

    def create_disk(self, vm_id, spec):
        """
        Adds a virtual disk to the virtual machine

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        :param spec: new virtual disk specification
        :type spec: dict
        """
        return self._post('vcenter/vm/{}/hardware/disk'.format(vm_id), json=spec)

    def get_disk(self, vm_id, disk_id):
        """
        Returns information about a virtual disk.

        :param vm_id: Virtual machine identifier.
        :type vm_id: string
        :param disk_id: Virtual disk identifier.
        :type disk_id: string
        """
        return self._get('vcenter/vm/{}/hardware/disk/{}'.format(vm_id, disk_id))

    def delete_disk(self, vm_id, disk_id):
        """
        Removes a virtual disk from the virtual machine.
        This operation does not destroy the VMDK file that backs the virtual disk.
        It only detaches the VMDK file from the virtual machine.
        Once detached, the VMDK file will not be destroyed when the virtual machine
        to which it was associated is deleted.

        :param vm_id: Virtual machine identifier.
        :type vm_id: string
        :param disk_id: Virtual disk identifier.
        :type disk_id: string
        """
        return self._delete('vcenter/vm/{}/hardware/disk/{}'.format(vm_id, disk_id))

    def connect_cdrom(self, vm_id, cdrom_id):
        """
        Connects a virtual CD-ROM device of a powered-on virtual machine to its backing.

        :param vm_id: Virtual machine identifier
        :type vm_id: string
        :param cdrom_id: Virtual CD-ROM device identifier.
        :type cdrom_id: string
        """
        return self._post('vcenter/vm/{}/hardware/cdrom/{}/connect'.format(vm_id, cdrom_id))

    def disconnect_cdrom(self, vm_id, cdrom_id):
        """
        Disconnects a virtual CD-ROM device of a powered-on virtual machine from its backing.

        :param vm_id: Virtual machine identifier.
        :type vm_id: string
        :param cdrom_id: Virtual CD-ROM device identifier.
        :type cdrom_id: string
        """
        return self._post('vcenter/vm/{}/hardware/cdrom/{}/disconnect'.format(vm_id, cdrom_id))

    def create_nic(self, vm_id, network_id):
        """
        Adds a virtual Ethernet adapter to the virtual machine.

        :param vm_id: Virtual machine identifier.
        :type vm_id: string
        :param network_id: Identifier of the network that backs the virtual Ethernet adapter.
        :type network_id: string
        """
        spec = {
            'backing': {
                'network': network_id,
                'type': 'DISTRIBUTED_PORTGROUP',
            },
            'start_connected': True,
        }
        return self._post('vcenter/vm/{}/hardware/ethernet'.format(vm_id), json=spec)

    def delete_nic(self, vm_id, nic_id):
        """
        Removes a virtual Ethernet adapter from the virtual machine.

        :param vm_id: Virtual machine identifier.
        :type vm_id: string
        :param nic_id: Virtual Ethernet adapter identifier.
        :type nic_id: string
        """
        return self._delete('vcenter/vm/{}/hardware/ethernet/{}'.format(vm_id, nic_id))

    def list_nics(self, vm_id):
        """
        Returns list of Ethernet adapters by virtual machine ID.

        :param vm_id: Virtual machine identifier.
        :type vm_id: string
        :rtype: List[Dict]
        """
        result = []
        for nic in self.list_nic_ids(vm_id):
            nic_payload = self.get_nic(vm_id, nic['nic'])
            nic_payload['nic'] = nic['nic']
            result.append(nic_payload)
        return result

    def list_nic_ids(self, vm_id):
        """
        Returns list of Ethernet adapter IDs by virtual machine ID.

        :param vm_id: Virtual machine identifier.
        :type vm_id: string
        """
        return self._get('vcenter/vm/{}/hardware/ethernet'.format(vm_id))

    def get_nic(self, vm_id, nic_id):
        """
        Returns information about a virtual Ethernet adapter.

        :param vm_id: Virtual machine identifier.
        :type vm_id: string
        :param nic_id: Virtual Ethernet adapter identifier.
        :type nic_id: string
        """
        return self._get('vcenter/vm/{}/hardware/ethernet/{}'.format(vm_id, nic_id))

    def connect_nic(self, vm_id, nic_id):
        """
        Connects a virtual Ethernet adapter of a powered-on virtual machine to its backing.

        :param vm_id: Virtual machine identifier.
        :type vm_id: string
        :param nic_id: Virtual Ethernet adapter identifier.
        :type nic_id: string
        """
        return self._post('vcenter/vm/{}/hardware/ethernet/{}/connect'.format(vm_id, nic_id))

    def disconnect_nic(self, vm_id, nic_id):
        """
        Disconnects a virtual Ethernet adapter of a powered-on virtual machine from its backing.

        :param vm_id: Virtual machine identifier.
        :type vm_id: string
        :param nic_id: Virtual Ethernet adapter identifier.
        :type nic_id: string
        """
        return self._post('vcenter/vm/{}/hardware/ethernet/{}/disconnect'.format(vm_id, nic_id))

    def list_libraries(self):
        return self._get('com/vmware/content/library')

    def list_library_items(self, library_id):
        params = {'library_id': library_id}
        return self._get('com/vmware/content/library/item', params=params)

    def get_library_item(self, library_item_id):
        return self._get('com/vmware/content/library/item/id:{}'.format(library_item_id))

    def get_template_library_item(self, library_item_id):
        return self._get('vcenter/vm-template/library-items/{}'.format(library_item_id))

    def list_all_templates(self):
        items = []
        for library_id in self.list_libraries():
            for library_item_id in self.list_library_items(library_id):
                library_item = self.get_library_item(library_item_id)
                if library_item['type'] == 'vm-template':
                    template = self.get_template_library_item(library_item_id)
                    items.append({
                        'library_item': library_item,
                        'template': template,
                    })
        return items

    def deploy_vm_from_template(self, library_item_id, spec):
        """
        Deploys a virtual machine as a copy of the source virtual machine
        template contained in the library item specified by library_item_id.

        :param library_item_id: identifier of the content library item containing the source virtual machine template to be deployed.
        :param spec: deployment specification
        :return: Identifier of the deployed virtual machine.
        :rtype: str
        """
        url = 'vcenter/vm-template/library-items/{}?action=deploy'.format(library_item_id)
        return self._post(url, json=spec)



import decimal
import math

from django.utils.functional import cached_property

from .base import BaseReportLine


class MoabReportLine(BaseReportLine):
    def __init__(self, line):
        self._parts = line.split('|')

    @cached_property
    def account(self):
        return self._parts[0].strip()

    @cached_property
    def user(self):
        return self._parts[5]

    @cached_property
    def cpu(self):
        return self.get_int(1)

    @cached_property
    def gpu(self):
        return self.get_int(2)

    @cached_property
    def ram(self):
        return self.get_int(3)

    @cached_property
    def duration(self):
        # convert seconds to minutes
        return int(math.ceil(self.get_int(4) / 60))

    @cached_property
    def charge(self):
        return decimal.Decimal(self._parts[6])

    @cached_property
    def node(self):
        return int(self._parts[7])

    def get_int(self, index):
        value = self._parts[index] or 0
        return int(value)
